#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This modules define the basic Actor object**

 Few methods are available:

    - add_external_constraint
    - add_external_dynamic_constraint
    - add_objective

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from ..general.optimisation.core import OptObject
from ..general.optimisation.elements import *

class Actor(OptObject):
    """
    **Description**

        Actor class is the basic class to model an actor. The basic actor
        is defined by its name and description.
        An actor is then defined by its constraints and objectives.

    **Attributes**

     - description : description as an Actor OptObject
    """

    def __init__(self, name, no_warn=True, verbose=True):
        OptObject.__init__(self, name=name, verbose=verbose)

        self.description = 'Actor OptObject'

    def add_actor_constraint(self, cst_name, exp):
        """
        Enable to add an external constraint linked with an actor

        :param cst_name: name of the constraint
        :param exp: expression of the constraint

        """
        setattr(self, cst_name, ActorConstraint(exp=exp,
                                                name=cst_name, parent=self))

    def add_actor_dynamic_constraint(self, cst_name, exp_t,
                                     t_range='for t in time.I'):
        """
        Enable to add an external dynamic constraint linked with an actor.
        A dynamic constraint changes over time

        :param cst_name: name of the constraint
        :param exp: expression of the constraint depending on the time
        :param t_range: expression of time for the constraint
        """
        setattr(self, cst_name, ActorDynamicConstraint(exp_t=exp_t,
                                                       name=cst_name,
                                                       t_range=t_range,
                                                       parent=self))

    def remove_actor_constraints(self, ext_cst_name_list=None):
        """
        Enable to remove an external constraint linked with an actor

        :param ext_cst_name_list: list of external constraint that would be
            removed
        """
        if not ext_cst_name_list:
            cst_dict = self.get_constraints_list()
            for key in cst_dict:
                if isinstance(cst_dict[key],
                              ActorConstraint or ActorDynamicConstraint):
                    cst_dict[key].deactivate_constraint()
                else:
                    Warning('Only external constraints can be removed')
        else:
            cst_dict = self.get_constraints_list()
            for cst_name in ext_cst_name_list:
                for key in cst_dict:
                    if cst_name == key:
                        if isinstance(cst_dict[key],
                                      ActorConstraint or
                                      ActorDynamicConstraint):
                            cst_dict[key].deactivate_constraint()
                        else:
                            Warning('Only external constraints can be removed')
                    else:
                        pass

    def add_objective(self, obj_name, exp):
        """
        Enable to add an objective linked with an actor

        :param obj_name: name of the objective
        :param exp: expression of the objective
        """
        setattr(self, obj_name, Objective(exp=exp, name=obj_name, parent=self))



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module includes the following utils for output data management**

 It contains the following methods:
    - save_energy_flows() : Save the optimisation results in a .csv file

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import csv

__docformat__ = "restructuredtext en"


def save_energy_flows(*nodes, file_name=None, sep='\t', decimal_sep='.'):
    """
    Save the optimisation results in a .csv file

    :param nodes: list of the nodes from which should be collected the data
    :param file_name: name of the file to save the data
    :param sep: separator for the data
    :param decimal_sep: separator dor the decimals of the data
    """

    if decimal_sep != '.' and decimal_sep != ',':
        raise ValueError("The decimal separator should be either a dot "
                         "or a comma but is {0}".format(decimal_sep))
    if file_name is None:
        file_name = 'energy_flows_results.csv'
    else:
        file_name += '.csv'

    time = getattr(nodes[0], 'time')

    energy_flows = [
        ['date'] + [date.to_pydatetime() for date in time.DATES] + ['hour']]
    for node in nodes:
        for energy_flow in node.get_flows:
            v = getattr(energy_flow, 'value')
            parent = getattr(energy_flow, 'parent')
            name = getattr(parent, 'name')

            if isinstance(v, list):
                # Using european format
                if decimal_sep == ',':
                    v_euro = [str(el).replace('.', ',')
                              if isinstance(el, float) else el for el in v]
                    energy_flows.append([name] + v_euro)
                else:
                    energy_flows.append([name] + v)
            elif isinstance(v, dict):
                if decimal_sep == ',':
                    v_values_euro = [str(el).replace('.', ',')
                                     if isinstance(el, float) else el for el
                                     in list(v.values())]
                    energy_flows.append([name] + v_values_euro)
                else:
                    energy_flows.append([name] + list(v.values()))
            else:
                pass

    with open(file_name, 'w', newline='') as f:
        writer = csv.writer(f, delimiter=sep)
        writer.writerows(zip(*energy_flows))



#! usr/bin/env python3
# -*- coding: utf-8 -*-

"""
**This module describes the Consumer actor**

 Few objectives and constraints are available.

 Objectives :
    - maximize_consumption
    - minimize_consumption
    - minimize_co2_consumption
    - minimize_consumption_costs

 Constraints :
    - energy_consumption_minimum
    - energy_consumption_maximum
    - power_consumption_minimum
    - power_consumption_maximum

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from .operator_actors import OperatorActor
from ...general.optimisation.elements import *
from ...energy.units.consumption_units import ConsumptionUnit
from ...energy.energy_nodes import EnergyNode

__docformat__ = "restructuredtext en"


class Consumer(OperatorActor):
    """
    **Description**

        Consumer class inherits from the the class OperatorActor. It enables
        one to model a consumer actor.

    """

    def __init__(self, name, operated_unit_list, operated_node_list=None,
                 verbose=True):
        OperatorActor.__init__(self, name=name,
                               operated_unit_type_tuple=(ConsumptionUnit,
                                                         EnergyNode),
                               operated_unit_list=operated_unit_list,
                               operated_node_list=operated_node_list,
                               verbose=verbose)

    # OBJECTIVES #
    # Energy objectives
    def minimize_consumption(self, obj_operated_unit_list=None, weight=1):
        """
        To create the objective in order to minimize the consumption of the
        consumer's units (all or part of them).

        :param obj_operated_unit_list: List of units on which the
            objective will be applied. Might be empty
        :param weight: Weight coefficient for the objective

        """
        final_operated_unit_list = self._check_operated_unit_list(
            obj_operated_unit_list)
        for operated_unit in final_operated_unit_list:
            operated_unit.minimize_consumption(weight=weight)

    def maximize_consumption(self, obj_operated_unit_list=None, weight=1):
        """
        To create the objective in order to maximize the consumption of the
        consumer's units (all or part of them).

        :param obj_operated_unit_list: List of units on which the objective
            will be applied. Might be empty
        :param weight: Weight coefficient for the objective

        """
        final_operated_unit_list = self._check_operated_unit_list(
            obj_operated_unit_list)
        for operated_unit in final_operated_unit_list:
            operated_unit.maximize_consumption(weight=weight)

    # Economic objectives
    def minimize_consumption_cost(self, obj_operated_unit_list=None, weight=1):
        """
        To create the objective in order to minimize the expenses due to the
        consumer's units (all or part of them).

        :param obj_operated_unit_list: List of units on which the objective
            will be applied. Might be empty
        :param weight: Weight coefficient for the objective

        """
        final_operated_unit_list = self._check_operated_unit_list(
            obj_operated_unit_list)
        for operated_unit in final_operated_unit_list:
            operated_unit.minimize_consumption_cost(weight=weight)

    # CO2 objectives
    def minimize_co2_consumption(self, obj_operated_unit_list=None, weight=1):
        """
        To create the objective in order to minimize the co2 emissions due to
        the consumer's units (all or part of them).

        :param obj_operated_unit_list: List of units on which the objective
            will be applied. Might be empty
        :param weight: Weight coefficient for the objective

        """
        final_operated_unit_list = self._check_operated_unit_list(
            obj_operated_unit_list)
        for operated_unit in final_operated_unit_list:
            operated_unit.minimize_co2_emissions(weight=weight)

    # CONSTRAINTS #
    # Energy constraints
    def energy_consumption_minimum(self, min_e_tot,
                                   cst_operated_unit_list=None):
        """
        To create the actor constraint of a minimum of energy consumption.

        :param min_e_tot: Minimum of the total energy consumption over the
            study period
        :param cst_operated_unit_list: List of units on which the constraint
            will be applied. Might be empty.

        """
        if isinstance(min_e_tot, (int, float)):
            final_operated_unit_list = self._check_operated_unit_list(
                cst_operated_unit_list)

            consumption_name_string = ''
            for consumption_unit in final_operated_unit_list:
                consumption_name_string += '_{}'.format(
                    consumption_unit.name)

            consumption_p_string = ''
            for consumption_unit in final_operated_unit_list[:-1]:
                consumption_p_string += '{}_p[t] + '.format(
                    consumption_unit.name)
            consumption_p_string += '{}_p[t] '.format(
                final_operated_unit_list[-1].name)

            cst_name = 'min_energy_conso{}'.format(consumption_name_string)
            exp = 'lpSum({0} for t in time.I) >= {1}'.format(
                consumption_p_string, min_e_tot)

            setattr(self, cst_name, ActorConstraint(exp=exp,
                                                    name=cst_name,
                                                    parent=self))

        else:
            raise TypeError('min_e_tot in energy_consumption_minimum '
                            'constraint should be an int or a float')

    def energy_consumption_maximum(self, max_e_tot,
                                   cst_operated_unit_list=None):
        """
        To create the actor constraint of a maximum of energy consumption.

        :param max_e_tot: Maximum of the total energy consumption over the
            study period
        :param cst_operated_unit_list: List of units on which the constraint
            will be applied. Might be empty.

        """
        if isinstance(max_e_tot, (int, float)):
            final_operated_unit_list = self._check_operated_unit_list(
                cst_operated_unit_list)

            consumption_name_string = ''
            for consumption_unit in final_operated_unit_list:
                consumption_name_string += '_{}'.format(
                    consumption_unit.name)

            consumption_p_string = ''
            for consumption_unit in final_operated_unit_list[:-1]:
                consumption_p_string += '{}_p[t] + '.format(
                    consumption_unit.name)
            consumption_p_string += '{}_p[t] '.format(
                final_operated_unit_list[-1].name)

            cst_name = 'max_energy_conso{}'.format(consumption_name_string)
            exp = 'lpSum({0} for t in time.I) <= {1}'.format(
                consumption_p_string, max_e_tot)

            setattr(self, cst_name, ActorConstraint(exp=exp,
                                                    name=cst_name,
                                                    parent=self))
        else:
            raise TypeError('min_e_tot in energy_consumption_minimum '
                            'constraint should be an int or a float')

    def power_consumption_minimum(self, min_p, time,
                                  cst_operated_unit_list=None):
        """
        To create the actor constraint of a minimum of power consumption.

        :param min_p: Minimum of the power consumption. May be an int,
            float or a list with the size of the period study
        :param time: period of the study
        :param cst_operated_unit_list: List of units on which the constraint
            will be applied. Might be empty.

        """
        final_operated_unit_list = self._check_operated_unit_list(
            cst_operated_unit_list)

        consumption_name_string = ''
        for consumption_unit in final_operated_unit_list:
            consumption_name_string += '_{}'.format(
                consumption_unit.name)

        consumption_p_string = ''
        for consumption_unit in final_operated_unit_list[:-1]:
            consumption_p_string += '{}_p[t] + '.format(consumption_unit.name)
        consumption_p_string += '{}_p[t]'.format(
            final_operated_unit_list[-1].name)

        cst_name = 'min_power_conso{}'.format(consumption_name_string)

        if isinstance(min_p, (int, float)):
            exp = '{0} >= {1}'.format(consumption_p_string, min_p)
            setattr(self, cst_name, ActorConstraint(exp=exp,
                                                    name=cst_name,
                                                    parent=self))

        elif isinstance(min_p, list):
            if len(min_p) != time.LEN:
                raise IndexError(
                    "Your minimal power should be the size of the time "
                    "period : {} but equals {}.".format(time.LEN,
                                                        len(min_p)))
            else:
                exp_t = '{0} >= {1}[t]'.format(consumption_p_string, min_p)
                setattr(self, cst_name, ActorDynamicConstraint(exp_t=exp_t,
                                                               name=cst_name,
                                                               t_range='for '
                                                                       't in '
                                                                 'time.I',
                                                               parent=self))
        else:
            raise TypeError('Your minimal power should be an int, '
                            'a float or a list.')

    def power_consumption_maximum(self, max_p, time,
                                  cst_operated_unit_list=None):
        """
        To create the actor constraint of a maximum of power consumption.

        :param max_p: Maximum of the power consumption. May be an int,
            float or a list with the size of the period study
        :param time: period of the study
        :param cst_operated_unit_list: List of units on which the constraint
            will be applied. Might be empty.

        """
        final_operated_unit_list = self._check_operated_unit_list(
            cst_operated_unit_list)

        consumption_name_string = ''
        for consumption_unit in final_operated_unit_list:
            consumption_name_string += '_{}'.format(
                consumption_unit.name)

        consumption_p_string = ''
        for consumption_unit in final_operated_unit_list[:-1]:
            consumption_p_string += '{}_p[t] + '.format(consumption_unit.name)
        consumption_p_string += '{}_p[t]'.format(
            final_operated_unit_list[-1].name)

        cst_name = 'max_power_conso{}'.format(consumption_name_string)

        if isinstance(max_p, (int, float)):
            exp = '{0} <= {1}'.format(consumption_p_string, max_p)
            setattr(self, cst_name, ActorConstraint(exp=exp,
                                                    name=cst_name,
                                                    parent=self))

        elif isinstance(max_p, list):
            if len(max_p) != time.LEN:
                raise IndexError(
                    "Your maximal power should be the size of the time "
                    "period : {} but equals {}.".format(time.LEN,
                                                        len(max_p)))
            else:
                exp_t = '{0} <= {1}[t]'.format(consumption_p_string, max_p)
                setattr(self, cst_name, ActorDynamicConstraint(exp_t=exp_t,
                                                               name=cst_name,
                                                               t_range='for t in '
                                                                 'time.I',
                                                               parent=self))
        else:
            raise TypeError('Your maximal power should be an int, '
                            'a float or a list.')



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module defines the energy units of OMEGAlpes. The production,
consumption and storage unit will inherit from it.**

 The energy_units module defines the basic attributes and methods of an
 energy unit in OMEGAlpes.

 The class EnergyUnit includes the following attributes and quantities:
    - p : instantaneous power of the energy unit (kW)
    - p_min : minimal power (kW)
    - p_max : maximal power (kW)
    - e_tot : total energy during the time period (kWh)
    - e_min : minimal energy of the unit (kWh)
    - e_max : maximal energy of the unit (kWh)
    - u : binary describing if the unit is operating or not at t (delivering or
      consuming P)

 This module also includes the classes:
 - FixedEnergyUnit: energy unit with a fixed power profile

 - VariableEnergyUnit: energy unit with a variable power profile

 - SquareEnergyUnit: energy unit with a defined square power profile,
 inheriting from VariableEnergyUnit.

 - ShiftableEnergyUnit: energy unit with a power profile that can be time
 shifted , inheriting from VariableEnergyUnit.

 - TriangleEnergyUnit: energy unit with a defined triangular power profile,
 inheriting from VariableEnergyUnit.

 - SawtoothEnergyUnit: energy unit with a defined sawtooth power profile,
 inheriting from VariableEnergyUnit.

 - SeveralEnergyUnit: Energy unit based on a fixed power curve enabling to
  multiply several times (nb_unit) the same power curve.

 - AssemblyUnit: an assembly unit has at least a production unit and a
 consumption unit and is using one or several energy types. It can also
 integrate reversible energy units. It inherits from OptObject and it is the
 parent class of ConversionUnit and ReversibleUnit.

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import warnings

import numpy as np
import pandas as pd
from pulp import LpBinary, LpInteger, LpContinuous

from ..io.poles import Epole
from ...general.optimisation.core import OptObject
from ...general.optimisation.elements import *

__docformat__ = "restructuredtext en"


class EnergyUnit(OptObject):
    """
        **Description**

        Module dedicated to the parent class (EnergyUnit) of :

         - production units
         - consumption units
         - storage units
    """

    def __init__(self, time, name, flow_direction='in', p=None, p_min=-1e+4,
                 p_max=1e+4, e_min=-1e6, e_max=1e6, starting_cost=None,
                 operating_cost=None, min_time_on=None, min_time_off=None,
                 max_ramp_up=None, max_ramp_down=None, co2_out=None,
                 availability_hours=None, energy_type=None,
                 no_warn=True, verbose=True):

        OptObject.__init__(self, name=name, description='Energy unit',
                           verbose=verbose)

        self.parent = None
        self.time = time  # Time unit
        self.energy_type = energy_type

        self.set_e_min = None
        self.set_e_max = None

        if isinstance(p_max, (int, float)):
            p_ub = max(0, p_max)  # Could be 0 when turn off
            max_e = p_ub * time.DT * time.LEN
        elif isinstance(p_max, list):
            p_ub = [max(0, p) for p in p_max]  # Could be 0 when turn off
            max_e = max(p_ub) * time.DT * time.LEN

        if isinstance(p_min, (int, float)):
            p_lb = min(0, p_min)  # Could be 0 when turn off
        elif isinstance(p_min, list):
            p_lb = [min(0, p) for p in p_min]  # Could be 0 when turn off

        self.p = Quantity(name='p',
                          description='instantaneous power of the energy unit',
                          value=p, lb=p_lb, ub=p_ub, vlen=time.LEN, unit='kW',
                          parent=self)

        self.e_tot = Quantity(name='e_tot',
                              description='total energy during the time '
                                          'period', lb=e_min, ub=e_max, vlen=1,
                              unit='kWh', parent=self)

        self.calc_e_tot = DefinitionConstraint(name='calc_e_tot', parent=self,
                                               exp='{0}_e_tot == time.DT * '
                                                   'lpSum({0}_p[t] for t in '
                                                   'time.I)'
                                               .format(self.name))

        if isinstance(self, VariableEnergyUnit):
            if max_e > e_max and not isinstance(self, ShiftableEnergyUnit):
                if not no_warn:
                    warnings.warn("Your EnergyUnit {0} won't be able to "
                                  "operate every time steps at its maximal "
                                  "power {1}.".format(name, max_e /
                                                      (time.DT * time.LEN)),
                                  UserWarning)

            # CONSTRAINTS
            self.u = Quantity(name='u',
                              description='indicates if the unit is operating '
                                          'at t',
                              vtype=LpBinary, vlen=time.LEN, parent=self)

            if isinstance(p_max, (int, float)):
                self.on_off_max = DefinitionDynamicConstraint(
                    exp_t='{0}_p[t] <= {0}_u[t] * {p_M}'.format(self.name,
                                                                p_M=p_max),
                    t_range='for t in time.I', name='on_off_max', parent=self)
            elif isinstance(p_max, list):
                self.on_off_max = DefinitionDynamicConstraint(
                    exp_t='{0}_p[t] <= {0}_u[t] * {p_M}[t]'.format(self.name,
                                                                   p_M=p_max),
                    t_range='for t in time.I', name='on_off_max', parent=self)

            if isinstance(p_min, (int, float)):
                self.on_off_min = DefinitionDynamicConstraint(
                    exp_t='{0}_p[t] >= {0}_u[t] * {p_m}'.format(self.name,
                                                                p_m=p_min),
                    t_range='for t in time.I', name='on_off_min',
                    parent=self)
            elif isinstance(p_min, list):
                self.on_off_min = DefinitionDynamicConstraint(
                    exp_t='{0}_p[t] >= {0}_u[t] * {p_m}[t]'.format(self.name,
                                                                   p_m=p_min),
                    t_range='for t in time.I', name='on_off_min',
                    parent=self)

        # Poles of the energy unit
        self.poles = {1: Epole(self.p, flow_direction, energy_type)}

        self._add_opt_parameters(starting_cost, operating_cost, co2_out)
        self._add_tech_constraints(min_time_on, min_time_off, max_ramp_up,
                                   max_ramp_down)
        self._add_actor_constraints(availability_hours)

    def _set_optobject_as_attr(self, optobject: OptObject,
                               attribute_name=None):
        if isinstance(optobject, OptObject):
            if attribute_name:
                self.__setattr__(attribute_name, optobject)
            else:
                self.__setattr__(optobject.name, optobject)

        else:
            raise TypeError('In the _set_optobject_as_attr method, '
                            'the object should '
                            'be an OptObject class and {0} is a '
                            '{1}'.format(optobject, type(optobject)))

    # Optional parameters
    def _add_opt_parameters(self, starting_cost, operating_cost, co2_out):
        self.start_up = None
        self.switch_off = None
        self.starting_cost = None
        self.operating_cost = None
        self.co2_emissions = None

        # Starting cost
        if starting_cost is not None:
            self.add_starting_cost(starting_cost)

        # Operating cost
        if operating_cost is not None:
            self.add_operating_cost(operating_cost)

        # CO2 emissions
        if co2_out is not None:
            self.add_co2_emissions(co2_out)

    def _add_start_up(self):
        time = self.time
        # Add a variable for start up
        self.start_up = Quantity(name='start_up',
                                 description='The EnergyUnit is '
                                             'starting :1 or not :0',
                                 vtype=LpBinary, vlen=time.LEN, parent=self)

        # When u[t] = 0 and u[t+1] = 1, start_up[t+1] = 1
        self.def_start_up = DefinitionDynamicConstraint(
            exp_t='{0}_u[t+1] - {0}_u[t] <= '
                  '{0}_start_up[t+1]'.format(self.name),
            t_range='for t in time.I[:-1]', name='def_start_up', parent=self)

        # Else start_up[t+1] = 0
        self.def_no_start_up = DefinitionDynamicConstraint(
            exp_t='{0}_start_up[t+1] <= ({0}_u[t+1] - {0}_u[t]'
                  ' + 1)/2'.format(self.name),
            t_range='for t in time.I[:-1]',
            name='def_no_start_up', parent=self)

        # Def initial start_up
        self.def_init_start_up = DefinitionConstraint(
            exp='{0}_start_up[0] == {0}_u[0]'.format(self.name),
            name='def_init_start_up', parent=self)

    def _add_switch_off(self):
        time = self.time
        if self.start_up is None:
            self._add_start_up()

        # Add a variable for switch off
        self.switch_off = Quantity(name='switch_off',
                                   description='The EnergyUnit is '
                                               'switching off :1 or not :0',
                                   vtype=LpBinary, vlen=time.LEN,
                                   parent=self)

        # When u[t} = 1 and u[t+1] = 0, switch_off[t+1] = 0
        self.def_switch_off = DefinitionDynamicConstraint(
            exp_t='{0}_switch_off[t+1] == {0}_start_up[t+1] '
                  '+ {0}_u[t] - {0}_u[t+1]'.format(self.name),
            t_range='for t in time.I[:-1]',
            name='def_switch_off', parent=self)

        # Set initial switch_off to 0
        self.def_init_switch_off = DefinitionConstraint(
            exp='{0}_switch_off[0] == 0'.format(self.name),
            name='def_init_switch_off', parent=self)

    def add_starting_cost(self, start_cost: float):
        """
        Add a starting cost associated to the energy unit based on the value
        start_cost.
        Each time the energy unit is starting (or restarting)

            i.e. not functioning time t and functioning time t+1
            ( When start_up[t+1] = 1 corresponding to u[t] = 0 and u[t+1] = 1)

        the start_cost value is added to the starting_costs.

        :param start_cost: float: cost corresponding to the start-up of the
            energy unit
        """
        if self.starting_cost is None:
            if self.start_up is None:
                self._add_start_up()

            # Adding starting cost
            self.starting_cost = Quantity(name='starting_cost',
                                          description='Dynamic cost for the '
                                                      'start'
                                                      ' of EnergyUnit',
                                          lb=0, vlen=self.time.LEN,
                                          parent=self)

            # Defining how the starting cost is calculated
            self.calc_start_cost = DefinitionDynamicConstraint(
                exp_t='{0}_starting_cost[t] == {1} * {0}_start_up[t]'.format(
                    self.name, start_cost),
                t_range='for t in time.I[:-1]', name='calc_start_cost',
                parent=self)

        else:
            raise ValueError("The EnergyUnit {} already has a "
                             "starting cost defined.".format(self.name))

    def add_operating_cost(self, operating_cost: float):
        """
        Add an operating cost associated to the energy unit based on the value
        operating_cost.
        For each time step the energy unit is running the operating_cost
        value is multiplied by the power production or consumption and
        added to the operating_costs.

        :param operating_cost: float: cost corresponding operation of the
            energy unit. To be multiplied by the power at each time step
        """
        if self.operating_cost is None:
            # Adding operating cost
            self.operating_cost = Quantity(name='operating_cost',
                                           description='Dynamic cost for the '
                                                       'operation '
                                                       'of the EnergyUnit',
                                           lb=0,
                                           vlen=self.time.LEN, parent=self)

            if isinstance(operating_cost, (int, float)):
                self.calc_operating_cost = DefinitionDynamicConstraint(
                    name='calc_operating_cost',
                    exp_t='{0}_operating_cost[t] == {1} * '
                          '{0}_p[t] * time.DT'.format(self.name,
                                                      operating_cost),
                    t_range='for t in time.I', parent=self)

            elif isinstance(operating_cost, list):
                if len(operating_cost) != self.time.LEN:
                    raise IndexError(
                        "Your operating cost should be the size of the time "
                        "period. The time period is of {0} and your operating "
                        "cost have a size of {1}".format(self.time.LEN,
                                                         len(operating_cost)))
                else:
                    self.calc_operating_cost = DefinitionDynamicConstraint(
                        name='calc_operating_cost',
                        exp_t='{0}_operating_cost[t] == {1}[t] * '
                              '{0}_p[t] * time.DT'.format(self.name,
                                                          operating_cost),
                        t_range='for t in time.I', parent=self)
            else:
                raise TypeError('The operating_cost should be an int, a float '
                                'or a list.')
        else:
            raise ValueError("The EnergyUnit {} already has an "
                             "operating cost defined.".format(self.name))

    def add_co2_emissions(self, co2_out: float):
        """
        Add an CO2 emissions associated to the energy unit based on the value
        co2_out.
        For each time step the energy unit is running the co2_out
        value is multiplied by the power production or consumption and
        added to the co2_emissions of the energy unit.

        :param co2_out: float: co2 emissions corresponding to the operation
            of the energy unit. To be multiplied by the power at each
            time step
        """
        if self.co2_emissions is None:
            # Adding CO2 emissions from production/consumption
            self.co2_emissions = Quantity(
                name='co2_emissions', description='Dynamic CO2 emissions '
                                                  'generated by the '
                                                  'EnergyUnit', lb=0,
                vlen=self.time.LEN, parent=self)

            if isinstance(co2_out, (int, float)):
                self.calc_co2_emissions = DefinitionDynamicConstraint(
                    exp_t='{0}_co2_emissions[t] == {1} * '
                          '{0}_p[t] * time.DT'.format(self.name, co2_out),
                    name='calc_co2_emissions', parent=self)
            elif isinstance(co2_out, list):
                if len(co2_out) != self.time.LEN:
                    raise IndexError(
                        "Your CO2 emissions (CO2_out) should be the size of "
                        "the time period. The time period is of {0} and your "
                        "CO2 emissions have a size of {1}".format(
                            self.time.LEN,
                            len(co2_out)))
                else:
                    self.calc_co2_emissions = DefinitionDynamicConstraint(
                        exp_t='{0}_co2_emissions[t] == {1}[t] * '
                              '{0}_p[t] * time.DT'.format(self.name, co2_out),
                        name='calc_co2_emissions', parent=self)
            else:
                raise TypeError('co2_out should be an int, a float or a list.')
        else:
            raise ValueError("The EnergyUnit {} already has CO2 "
                             "emissions defined.".format(self.name))

    # Technical constraints
    def _add_tech_constraints(self, min_time_on, min_time_off, max_ramp_up,
                              max_ramp_down):
        self.set_max_ramp_up = None
        self.set_max_ramp_down = None
        self.set_min_up_time = None
        self.set_min_down_time = None

        # Adding a maximal ramp up
        if max_ramp_up is not None:
            self.add_max_ramp_up(max_ramp_up)

        # Adding a maximal ramp down
        if max_ramp_down is not None:
            self.add_max_ramp_down(max_ramp_down)

        # Adding a minimum time on
        if min_time_on is not None:
            self.add_min_time_on(min_time_on)

        # Adding a minimum time off
        if min_time_off is not None:
            self.add_min_time_off(min_time_off)

    def add_max_ramp_up(self, max_ramp_up: float):
        """
        Add a maximal ramp value between two consecutive power values
        increasing

        :param max_ramp_up: float: maximal ramp value between two consecutive
            power values increasing
        """
        if self.set_max_ramp_up is None:
            self.set_max_ramp_up = TechnicalDynamicConstraint(
                exp_t='{0}_p[t+1] - {0}_p[t] <= {1}'.format(self.name,
                                                            max_ramp_up),
                t_range='for t in time.I[:-1]', name='set_max_ramp_up',
                parent=self)
        else:
            raise ValueError("The EnergyUnit {} already has a maximal "
                             "ramp up defined.".format(self.name))

    def add_max_ramp_down(self, max_ramp_down: float):
        """
        Add a maximal ramp value between two consecutive power values
        decreasing

        :param max_ramp_down: float: maximal ramp value between two consecutive
            power values decreasing
        """
        if self.set_max_ramp_down is None:
            self.set_max_ramp_down = TechnicalDynamicConstraint(
                exp_t='{0}_p[t] - {0}_p[t+1] <= {1}'.format(self.name,
                                                            max_ramp_down),
                t_range='for t in time.I[:-1]', name='set_max_ramp_down',
                parent=self)
        else:
            raise ValueError("The EnergyUnit {} already has a maximal "
                             "ramp down".format(self.name))

    def add_min_time_on(self, min_time_on: float):
        """
        Add a minimal time during which the energy unit should function once
        it is started-up

        :param min_time_on: float: minimal time during which the energy unit
            should function once it is started-up
        """
        if self.set_min_up_time is None:
            if self.start_up is None:
                self._add_start_up()

            # When the unit starts, it should be on during min_time_on
            self.set_min_up_time = TechnicalDynamicConstraint(
                exp_t='{0}_u[t] >= lpSum({0}_start_up[i] for i in range('
                      'max(t - {1} + 1, 0), t))'.format(self.name,
                                                        min_time_on),
                t_range='for t in time.I', name='set_min_up_time', parent=self)
        else:
            raise ValueError("The EnergyUnit {} already has a "
                             "minimum time on.".format(self.name))

    def add_min_time_off(self, min_time_off: float):
        """
        Add a minimal time during which the energy unit has to remain off
        once it is switched off

        :param min_time_off: float: minimal time during which the energy unit
            has to remain off once it is switched off
        """
        if self.set_min_down_time is None:
            if self.switch_off is None:
                self._add_switch_off()
            # When the unit switches off, it should be off during min_time_off
            self.set_min_down_time = TechnicalDynamicConstraint(
                exp_t='1 - {0}_u[t] >= lpSum({0}_switch_off[i] for i in range('
                      'max(t - {1} + 1, 0), t))'.format(self.name,
                                                        min_time_off),
                t_range='for t in time.I', name='set_min_down_time',
                parent=self)

        else:
            raise ValueError("The EnergyUnit {} already has a "
                             "minimum time down.".format(self.name))

    # Actor constraints
    def _add_actor_constraints(self, availability_hours):
        self.set_availability = None

        # Adding a number of available hours of operation
        if availability_hours is not None:
            self.add_availability(availability_hours)

    def add_availability(self, av_hours: int):
        """
        Add a number of hours of availability of the energy unit during the
        study period

        :param av_hours: int: number of hours of availability of the energy
            unit during the study period
        """
        if self.set_availability is None:
            self.set_availability = ActorConstraint(
                exp='lpSum({dt} * {name}_u[t] for t in time.I) <= '
                    '{av_h}'.format(dt=self.time.DT, name=self.name,
                                    av_h=av_hours),
                name='set_availability', parent=self)
        else:
            raise ValueError("The EnergyUnit {} already has hours of "
                             "availability defined.".format(self.name))

    def set_operating_time_range(self, operating_time_range=[[str, str]]):
        """
        Add a range of hours during which the energy unit can be operated.
        The final time should be greater than the initial time within a time
        range, except when the final time is '00:00'.

        example: set_operating_time_range([['10:00', '12:00'], ['14:00',
        '17:00']])

        :param operating_time_range: list of lists of strings in the format
        HH:MM [[first hour operating: str,
        hour to stop (not operating): str], [second hour operating:
        str, hour to stop (not operating): str], etc]
        """

        # get the index of each hour for each range

        for time_range in range(len(operating_time_range)):
            (init_h_str, init_min_str) = operating_time_range[
                time_range][0].split(":")
            (final_h_str, final_min_str) = operating_time_range[
                time_range][1].split(":")

            init_h = int(init_h_str)
            init_min = int(init_min_str)
            final_h = int(final_h_str)
            final_min = int(final_min_str)

            start_time = datetime.datetime(year=self.time.get_days[0].year,
                                           month=self.time.get_days[
                                               0].month,
                                           day=self.time.get_days[0].day,
                                           hour=init_h, minute=init_min)
            start_index = self.time.get_index_for_date(date=start_time)

            if final_h != 0 or final_min != 0:
                end_time = datetime.datetime(year=self.time.get_days[0].year,
                                             month=self.time.get_days[0].month,
                                             day=self.time.get_days[0].day,
                                             hour=final_h, minute=final_min)
                end_index = self.time.get_index_for_date(date=end_time)
            else:
                end_index = 24 * 1 / self.time.DT

            operating_time_range[time_range] = [start_index, end_index]
            print('operating time range index {0}: '.format(time_range + 1),
                  operating_time_range[time_range])

        # First non-operating period before the first operating time range
        if operating_time_range[0][0] == 0:
            pass
        else:
            final_time = self.time.get_date_for_index(
                operating_time_range[0][0] - 1)

            set_start_time_range = DailyDynamicConstraint(
                exp_t='{name}_u[t] == 0'.format(name=self.name),
                time=self.time,
                init_time='00:00',
                final_time="{0}:{1}".format(final_time.hour,
                                            final_time.minute),
                name='set_operating_init_time_range_{}'.format(
                    operating_time_range[0][0]), parent=self)
            setattr(self, 'set_start_time_range_{}'.format(
                operating_time_range[0][0]), set_start_time_range)

        # Following non-operating periods between the given time range(s)
        if len(operating_time_range) != 1:
            set_time_range = []
            for i in range(1, len(operating_time_range)):
                init_time = self.time.get_date_for_index(
                    operating_time_range[i - 1][1])
                end_time = self.time.get_date_for_index(
                    operating_time_range[i][0] - 1)

                set_time_range.append(DailyDynamicConstraint(
                    exp_t='{name}_u[t] == 0'.format(name=self.name),
                    time=self.time,
                    init_time="{0}:{1}".format(init_time.hour,
                                               init_time.minute),
                    final_time="{0}:{1}".format(end_time.hour,
                                                end_time.minute),
                    name='set_time_range_{}_{}'.format(
                        operating_time_range[i - 1][1],
                        operating_time_range[i][0]),
                    parent=self))
                setattr(self,
                        'set_time_range_{}_{}'.format(
                            operating_time_range[i - 1][1],
                            operating_time_range[i][0]),
                        set_time_range[i - 1])

        # Last non-operating period after the last operating time range
        if operating_time_range[-1][1] == int(24 * 1 / self.time.DT):
            pass
        else:
            init_time = self.time.get_date_for_index(
                operating_time_range[-1][1])
            end_time = self.time.get_date_for_index(
                int(24 * 1 / self.time.DT - 1))
            set_end_time_range = DailyDynamicConstraint(
                exp_t='{name}_u[t] == 0'.format(name=self.name),
                time=self.time,
                init_time="{0}:{1}".format(init_time.hour, init_time.minute),
                final_time="{0}:{1}".format(end_time.hour, end_time.minute),
                name='set_operating_final_time_range_{}'.format(
                    operating_time_range[-1][1]),
                parent=self)
            setattr(self,
                    'set_end_time_range_{}'.format(operating_time_range[-1][
                                                       1]),
                    set_end_time_range)

    def add_operating_time_range(self, operating_time_range=[[int, int]]):
        """
        DEPRECATED : please use set_operating_time_range
        Add a range of hours during which the energy unit can be operated

        example: [[10, 12], [14, 17]]

        :param operating_time_range: [[first hour of functioning : int,
            hour to stop (not in functioning): int]]
        """

        import warnings
        deprecated_msg = "DEPRECATED: add_operating_time_range is deprecated" \
                         " and will be removed in omegalpes version 0.4 " \
                         "Please use set_operating_time_range that allows to" \
                         " use hour and minutes in the operating_time_range " \
                         "definition"
        warnings.warn(deprecated_msg, DeprecationWarning)

        if operating_time_range[0][0] == 0:
            pass

        else:
            set_start_time_range = HourlyDynamicConstraint(
                exp_t='{name}_u[t] == 0'.format(name=self.name),
                time=self.time,
                init_h=0,
                final_h=operating_time_range[0][0] - 1,
                name='set_operating_init_time_range_{}'.format(
                    operating_time_range[0][0]),
                parent=self)
            setattr(self, 'set_start_time_range_{}'.format(
                operating_time_range[0][0]),
                    set_start_time_range)

        if len(operating_time_range) != 1:
            set_time_range = []
            for i in range(1, len(operating_time_range)):
                set_time_range.append(HourlyDynamicConstraint(
                    exp_t='{name}_u[t] == 0'.format(name=self.name),
                    time=self.time,
                    init_h=operating_time_range[i - 1][1],
                    final_h=operating_time_range[i][0] - 1,
                    name='set_time_range_{}_{}'.format(
                        operating_time_range[i - 1][1],
                        operating_time_range[i][0]),
                    parent=self))
                setattr(self,
                        'set_time_range_{}_{}'.format(
                            operating_time_range[i - 1][1],
                            operating_time_range[i][0]),
                        set_time_range[i - 1])

        if operating_time_range[-1][1] == 23:
            pass
        else:
            set_end_time_range = HourlyDynamicConstraint(
                exp_t='{name}_u[t] == 0'.format(name=self.name),
                time=self.time,
                init_h=operating_time_range[-1][1],
                final_h=23,
                name='set_operating_final_time_range_{}'.format(
                    operating_time_range[0][1]),
                parent=self)
            setattr(self,
                    'set_end_time_range_{}'.format(operating_time_range[0][1]),
                    set_end_time_range)

    def set_energy_limits_on_time_period(self, e_min=0, e_max=None,
                                         start='YYYY-MM-DD HH:MM:SS',
                                         end='YYYY-MM-DD HH:MM:SS',
                                         period_index=None):
        """
        Add an energy limit during a defined time period

        :param e_min: Minimal energy set during the time period (int or float)
        :param e_max: Maximal energy set during the time period (int or float)
        :param start: Date of start of the time period  YYYY-MM-DD HH:MM:SS (
            str)
        :param end: Date of end of the time period   YYYY-MM-DD HH:MM:SS (str)
        """
        if period_index is None:
            if start == 'YYYY-MM-DD HH:MM:SS':
                index_start = ''
            else:
                index_start = self.time.get_index_for_date(start)

            if end == 'YYYY-MM-DD HH:MM:SS':
                index_end = ''
            else:
                index_end = self.time.get_index_for_date(end)

            period_index = 'time.I[{start}:{end}]'.format(start=index_start,
                                                          end=index_end)

        if e_min != 0:
            self.set_e_min = TechnicalConstraint(
                exp='time.DT * lpSum({0}_p[t] for t in {1}) '
                    '>= {2}'.format(self.name, period_index, e_min),
                name='set_e_min', parent=self)

        if e_max is not None:
            if e_max > self.e_tot.ub:
                self.e_tot.ub = e_max

            self.set_e_max = TechnicalConstraint(
                exp='time.DT * lpSum({0}_p[t] for t in {1}) '
                    '<= {2}'.format(self.name, period_index, e_max),
                name='set_e_max', parent=self)

    # OBJECTIVES #
    def minimize_starting_cost(self, weight=1):
        """
        Objective to minimize the starting costs

        :param weight: Weight coefficient for the objective
        """
        if self.starting_cost is not None:
            self.min_start_cost = Objective(name='min_start_cost',
                                            exp='lpSum({0}_starting_cost[t] '
                                                'for t '
                                                'in time.I)'
                                            .format(self.name), weight=weight,
                                            parent=self)
        else:
            raise ValueError("You should add a starting cost before trying "
                             "to minimize_starting_cost.")

    def minimize_operating_cost(self, weight=1):
        """
        Objective to minimize the operating costs

        :param weight: Weight coefficient for the objective
        """
        if self.operating_cost is not None:
            self.min_operating_cost = Objective(
                name='min_operating_cost',
                exp='lpSum({}_operating_cost[t] for t in time.I)'.format(
                    self.name),
                weight=weight,
                parent=self)
        else:
            raise ValueError("You should add an operating cost before trying "
                             "to minimize_operating_cost.")

    def minimize_costs(self, weight=1):
        """
        Objective to minimize the costs (starting and operating costs)

        :param weight: Weight coefficient for the objective
        """
        if self.starting_cost is not None:
            self.minimize_starting_cost(weight)

        if self.operating_cost is not None:
            self.minimize_operating_cost(weight)

    def minimize_energy(self, weight=1):
        """
        Objective to minimize the energy of the energy unit

        :param weight: Weight coefficient for the objective
        """
        self.min_energy = Objective(name='min_energy',
                                    exp='lpSum({0}_p[t] for t in time.I)'
                                    .format(self.name), weight=weight,
                                    parent=self)

    def minimize_time_of_use(self, weight=1):
        """
        Objective to minimize the time of running of the energy unit

        :param weight: Weight coefficient for the objective
        """
        self.min_time_of_use = Objective(name='min_time_of_use',
                                         exp='lpSum({0}_u[t] for t in time.I)'
                                         .format(self.name), weight=weight,
                                         parent=self)

    def minimize_co2_emissions(self, weight=1):
        """
        Objective to minimize the co2 emissions of the energy unit

        :param weight: Weight coefficient for the objective
        """
        self.min_co2_emissions = Objective(name='min_CO2_emissions',
                                           exp='lpSum({0}_co2_emissions[t] '
                                               'for t in time.I)'.format(
                                               self.name),
                                           weight=weight, parent=self)

    def minimize_exergy_destruction(self, weight=1):
        """ This is the main objective of any exergetic optimization. """
        if hasattr(self, 'exergy_dest'):
            self.min_exergy_dest = Objective(name='min_exergy_destruction',
                                             exp='lpSum({0}_exergy_dest[t] '
                                                 'for t in time.I)'.
                                             format(self.name), weight=weight,
                                             parent=self)
        else:
            raise ValueError("You should initialize exergy calculations "
                             "on the unit named {0} before trying to "
                             "minimize its exergy destruction."
                             .format(self.name))

    def minimize_exergy(self, energy_unit=None, weight=1):
        """ Alternate objective of exergy optimization that may be
        interesting in some cases. """
        if hasattr(self, 'exergy'):
            self.min_exergy = Objective(name='min_exergy',
                                             exp='lpSum({0}_exergy[t] '
                                                 'for t in time.I)'.
                                             format(energy_unit.name), weight=weight,
                                             parent=energy_unit)
        else:
            raise ValueError("You should initialize exergy calculations "
                             "on the unit named {0} before trying to "
                             "minimize its exergy."
                             .format(energy_unit.name))


class FixedEnergyUnit(EnergyUnit):
    """
    **Description**

        Energy unit with a fixed power profile.

    **Attributes**

        * p : instantaneous power known by advance (kW)
        * energy_type : type of energy ('Electrical', 'Heat', ...)

    """

    def __init__(self, time, name: str, p: list or dict or pd.DataFrame,
                 flow_direction='in',
                 starting_cost=None, operating_cost=None, co2_out=None,
                 energy_type=None, verbose=True):
        if p is None:
            # FixedEnergyUnits should have data for the power values
            raise TypeError(
                "You have to define the power profile (p) for the "
                "FixedEnergyUnit !")

        # p may be a list, a dict or a dataframe (1 column only)
        if isinstance(p, list):
            e_tot = sum(p) * time.DT
            p_min = min(p)
            p_max = max(p)

        elif isinstance(p, dict):
            # Checking the length of the dictionary corresponds to the length
            #  of the time unit
            if len(time.DATES) == len(p):
                e_tot = sum(p.values()) * time.DT
                p_min = min(p.values())
                p_max = max(p.values())

        elif isinstance(p, pd.DataFrame):
            # Checking the dataframe is composed of only one column :
            # - dates for the indexes
            # - power values for the column
            if len(p.columns) != 1:
                raise ValueError('You should have only one columns for the '
                                 'dataframe of the power values of the '
                                 'FixedEnergyUnit {0} and you have {1} '
                                 'columns'.format(name, len(p.columns)))

            # Checking that the dates of the values correspond to the date of
            #  the study based on the time unit
            if len(time.DATES) == len(p.index):
                for i in range(len(time.DATES)):
                    if not time.DATES[i] == p.index[i]:
                        raise ValueError(
                            'The FixedEnergyUnit {0} does not operate on the '
                            'same time of the model: \n'
                            'dates for the model time: {1} \n'
                            'dates for the power values: {2}'.format(name,
                                                                     time.DATES,
                                                                     p.index))
            else:
                raise ValueError(
                    'The FixedEnergyUnit {0} does not operate on the '
                    'same time of the model: \n'
                    'dates for the model time: {1} \n'
                    'dates for the power values: {2}'.format(name,
                                                             time.DATES,
                                                             p.index))

            # convert the dataframe into a list of power values to
            # calculate e_tot
            label = p.columns.values
            p = p[label[0]].tolist()
            e_tot = sum(p) * time.DT
            p_min = min(p)
            p_max = max(p)

        else:
            raise TypeError(
                "The power profile (p) for the FixedEnergyUnit should be a "
                "list, a dictionary or a dataframe")

        EnergyUnit.__init__(self, time=time, name=name, p=p, p_min=p_min,
                            p_max=p_max, e_min=e_tot, e_max=e_tot,
                            flow_direction=flow_direction,
                            starting_cost=starting_cost,
                            operating_cost=operating_cost, min_time_on=None,
                            min_time_off=None, max_ramp_up=None,
                            max_ramp_down=None, co2_out=co2_out,
                            availability_hours=None, energy_type=energy_type,
                            verbose=verbose, no_warn=True)


class VariableEnergyUnit(EnergyUnit):
    def __init__(self, time, name, flow_direction='in', p_min=-1e+4,
                 p_max=1e+4, e_min=-1e6, e_max=1e6, starting_cost=None,
                 operating_cost=None, min_time_on=None, min_time_off=None,
                 max_ramp_up=None, max_ramp_down=None, co2_out=None,
                 availability_hours=None, energy_type=None,
                 verbose=True, no_warn=True):
        EnergyUnit.__init__(self, time, name, flow_direction=flow_direction,
                            p=None, p_min=p_min, p_max=p_max, e_min=e_min,
                            e_max=e_max, starting_cost=starting_cost,
                            operating_cost=operating_cost,
                            min_time_on=min_time_on,
                            min_time_off=min_time_off, max_ramp_up=max_ramp_up,
                            max_ramp_down=max_ramp_down, co2_out=co2_out,
                            availability_hours=availability_hours,
                            energy_type=energy_type,
                            verbose=verbose, no_warn=no_warn)


class SquareEnergyUnit(VariableEnergyUnit):
    def __init__(self, time, name, p_square, n_square, t_between_sq,
                 t_square=1, flow_direction='in', starting_cost=None,
                 operating_cost=None, co2_out=None, energy_type=None,
                 verbose=True, no_warn=True):
        """

        :param time:
        :param name:
        :param p_square: Power of the square
        :param n_square: Number of squares
        :param t_square: Duration of a square [h]
        :param t_between_sq: Duration between squares [h]
        :param flow_direction:
        :param e_min:
        :param e_max:
        :param starting_cost:
        :param operating_cost:
        :param min_time_on:
        :param min_time_off:
        :param max_ramp_up:
        :param max_ramp_down:
        :param co2_out:
        :param availability_hours:
        :param energy_type:
        """
        if not isinstance(t_square, int):
            raise TypeError('t_squre should be an integer, but is '
                            'a {}'.format(type(t_square)))

        energy = n_square * t_square * p_square

        if n_square == 1:
            min_time_on = None
            min_time_off = None
            av_h = t_square
        else:
            min_time_on = t_square
            min_time_off = t_between_sq
            av_h = None

        VariableEnergyUnit.__init__(self, time, name,
                                    flow_direction=flow_direction,
                                    p_min=p_square, p_max=p_square,
                                    e_min=energy,
                                    e_max=energy, starting_cost=starting_cost,
                                    operating_cost=operating_cost,
                                    min_time_on=min_time_on,
                                    min_time_off=min_time_off,
                                    max_ramp_up=None,
                                    max_ramp_down=None, co2_out=co2_out,
                                    availability_hours=av_h,
                                    energy_type=energy_type,
                                    verbose=verbose,
                                    no_warn=no_warn)


class ShiftableEnergyUnit(VariableEnergyUnit):
    """
    **Description**

        EnergyUnit with shiftable power profile.

    **AttributsAttributes**

        * power_values : power profile to shift (kW)
        * mandatory : indicates if the power is mandatory (True) or not (False)
        * starting_cost : cost of the starting of the EnergyUnit
        * operating_cost : cost of the operation (€/kW)
        * energy_type : type of energy ('Electrical', 'Heat', ...)

    """

    def __init__(self, time, name: str, flow_direction, power_values: list,
                 mandatory=True, co2_out=None, starting_cost=None,
                 operating_cost=None, energy_type=None,
                 verbose=True):
        # Crop the power profile
        while power_values[0] == 0:
            power_values = power_values[1:]
        while power_values[-1] == 0:
            power_values = power_values[:-1]

        # Works if all values are strictly positives
        epsilon = 0.00001 * min(p > 0 for p in power_values)
        power_profile = [max(epsilon, p) for p in power_values]

        e_max = sum(power_profile) * time.DT

        if mandatory:
            e_min = e_max
        else:
            e_min = 0

        p_min = min(power_profile)
        p_max = max(power_profile)

        VariableEnergyUnit.__init__(self, time, name=name,
                                    flow_direction=flow_direction,
                                    p_min=p_min, p_max=p_max, e_min=e_min,
                                    e_max=e_max, starting_cost=starting_cost,
                                    operating_cost=operating_cost,
                                    min_time_on=None, min_time_off=None,
                                    max_ramp_up=None, max_ramp_down=None,
                                    co2_out=co2_out, availability_hours=None,
                                    energy_type=energy_type,
                                    verbose=verbose,
                                    no_warn=True)

        self._add_start_up()

        self.power_values = Quantity(name='power_values', opt=False,
                                     value=power_values, parent=self)

        for i, _ in enumerate(power_values):
            cst_name = 'def_{}_power_value'.format(i)

            exp_t = "{0}_p[t] >= {0}_power_values[{1}] * " \
                    "{0}_start_up[t-{1}]".format(self.name, i)

            cst = DefinitionDynamicConstraint(name=cst_name, exp_t=exp_t,
                                              t_range="for t in time.I[{"
                                                      "}:-1]".format(
                                                  i),
                                              parent=self)
            setattr(self, cst_name, cst)


class TriangleEnergyUnit(ShiftableEnergyUnit):
    def __init__(self, time, name, flow_direction, p_peak, alpha_peak,
                 t_triangle: list, mandatory=True, starting_cost=None,
                 operating_cost=None, co2_out=None, energy_type=None,
                 verbose=True):

        if not isinstance(t_triangle, int):
            raise TypeError('t_triangle should be an integer, but is '
                            'a {}'.format(type(t_triangle)))

        t_peak = alpha_peak * t_triangle

        if alpha_peak == 0:
            ramp_1 = 0
            ramp_2 = - p_peak / (t_triangle - t_peak)
        elif alpha_peak == 1:
            ramp_1 = p_peak / t_peak
            ramp_2 = 0
        else:
            ramp_1 = p_peak / t_peak
            ramp_2 = - p_peak / (t_triangle - t_peak)

        t = np.arange(0, t_triangle)

        triangle_profile = np.piecewise(
            t, [t <= t_peak - 1, (t_peak - 1 < t) & (t < t_peak), t_peak <= t],
            [lambda t: ramp_1 * (2 * t + 1) / 2,
             lambda t: (ramp_1 * t * (t_peak - t)
                        + ramp_2 * (t + 1 - t_triangle) * (t + 1 - t_peak)
                        + p_peak) / 2,
             lambda t: ramp_2 * (2 * t + 1 - 2 * t_triangle) / 2])

        triangle_profile = [float(p) for p in triangle_profile]

        ShiftableEnergyUnit.__init__(self, time, name,
                                     flow_direction=flow_direction,
                                     power_values=list(triangle_profile),
                                     mandatory=mandatory, co2_out=co2_out,
                                     starting_cost=starting_cost,
                                     operating_cost=operating_cost,
                                     energy_type=energy_type,
                                     verbose=verbose,
                                     no_warn=True)


class SawtoothEnergyUnit(ShiftableEnergyUnit):
    def __init__(self, time, name, flow_direction, p_peak, p_low, alpha_peak,
                 t_triangle, t_sawtooth, mandatory=True, starting_cost=None,
                 operating_cost=None, co2_out=None, energy_type=None,
                 verbose=True):
        if not isinstance(t_triangle, int):
            raise TypeError('t_triangle should be an integer, but is '
                            'a {}'.format(type(t_triangle)))
        if not isinstance(t_sawtooth, int):
            raise TypeError('t_sawtooth should be an integer, but is '
                            'a {}'.format(type(t_sawtooth)))

        t_peak = alpha_peak * t_triangle

        if alpha_peak == 0:
            ramp_1 = 0
            ramp_2 = - p_peak / (t_triangle - t_peak)
        elif alpha_peak == 1:
            ramp_1 = p_peak / t_peak
            ramp_2 = 0
        else:
            ramp_1 = p_peak / t_peak
            ramp_2 = - p_peak / (t_triangle - t_peak)

        t = np.arange(0, t_triangle)

        triangle_profile = np.piecewise(
            t, [t <= t_peak - 1, (t_peak - 1 < t) & (t < t_peak), t_peak <= t],
            [lambda t: ramp_1 * (2 * t + 1) / 2,
             lambda t: (ramp_1 * t * (t_peak - t)
                        + ramp_2 * (t + 1 - t_triangle) * (t + 1 - t_peak)
                        + p_peak) / 2,
             lambda t: ramp_2 * (2 * t + 1 - 2 * t_triangle) / 2])

        for espace in range(t_sawtooth - 2 * t_triangle + 1):
            if (t_sawtooth - t_triangle) % (t_triangle + espace) == 0:
                N = int((t_sawtooth + espace) / (t_triangle + espace))
                break

        triangle = [float(P) for P in triangle_profile]

        sawtooth_profile = triangle + (N - 1) * (espace * [p_low] + triangle)

        ShiftableEnergyUnit.__init__(self, time, name,
                                     flow_direction=flow_direction,
                                     power_values=list(sawtooth_profile),
                                     mandatory=mandatory, co2_out=co2_out,
                                     starting_cost=starting_cost,
                                     operating_cost=operating_cost,
                                     energy_type=energy_type,
                                     verbose=verbose,
                                     no_warn=True)


class SeveralEnergyUnit(VariableEnergyUnit):
    """
    **Description**

        Energy unit based on a fixed power curve enabling to multiply
        several times (nb_unit) the same power curve.

        Be careful, if imaginary == True, the solution may be imaginary as
        nb_unit can be continuous. The accurate number of the power unit
        should be calculated later

    **Attributes**

        * fixed_power : fixed power curve

    """

    def __init__(self, time, name, fixed_power, p_min=1e-5, p_max=1e+5,
                 imaginary=False, e_min=0, e_max=1e6, nb_unit_min=0,
                 nb_unit_max=None, flow_direction='in', starting_cost=None,
                 operating_cost=None, max_ramp_up=None, max_ramp_down=None,
                 co2_out=None, energy_type=None,
                 verbose=True, no_warn=True):
        VariableEnergyUnit.__init__(self, time=time, name=name,
                                    flow_direction=flow_direction,
                                    p_min=p_min, p_max=p_max, e_min=e_min,
                                    e_max=e_max, starting_cost=starting_cost,
                                    operating_cost=operating_cost,
                                    min_time_on=None, min_time_off=None,
                                    max_ramp_up=max_ramp_up,
                                    max_ramp_down=max_ramp_down,
                                    co2_out=co2_out, availability_hours=None,
                                    energy_type=energy_type,
                                    verbose=verbose,
                                    no_warn=no_warn)

        self.power_curve = Quantity(name='power_curve', opt=False,
                                    value=fixed_power, vlen=time.LEN,
                                    parent=self)

        if imaginary:
            self.nb_unit = Quantity(name='nb_unit', opt=True,
                                    vtype=LpContinuous,
                                    lb=nb_unit_min, ub=nb_unit_max, vlen=1,
                                    parent=self)
            warnings.warn(
                'The solution may be imaginary as nb_unit is continuous')
        else:
            self.nb_unit = Quantity(name='nb_unit', opt=True, vtype=LpInteger,
                                    lb=nb_unit_min, ub=nb_unit_max, vlen=1,
                                    parent=self)

        self.calc_power_with_nb_unit_cst = DefinitionDynamicConstraint(
            exp_t='{0}_p[t] == {0}_nb_unit * {0}_power_curve[t]'.format(
                self.name), name='calc_power_with_nb_unit',
            t_range='for t in time.I', parent=self)


class AssemblyUnit(OptObject):
    """
    **Description**

        Simple Assembly unit: assembly units has at least a production unit
        and a consumption unit and is using one or several energy types.
        It  can also integrate reversible energy units. It inherits from
        OptObject and it  is the parent class of ConversionUnit and
        ReversibleUnit.

    **Attributes**

     * time: TimeUnit describing the studied time period
     * prod_units: list of the production units in the assembly unit.
     * cons_units: list of the consumption units in the assembly unit.
     * rev_units: list of the reversible units in the assembly unit.
     * poles: dictionary of the poles of the assembly unit

    """

    def __init__(self, time, name, prod_units=None, cons_units=None,
                 rev_units=None, verbose=True):
        OptObject.__init__(self, name=name, description='Assembly unit',
                           verbose=verbose)

        from .production_units import ProductionUnit
        from .consumption_units import ConsumptionUnit
        from .reversible_units import ReversibleUnit

        self.time = time
        self.prod_units = []  # Initialize an empty list for the
        # production units
        self.cons_units = []  # Initialize an empty list for the consumption
        # units
        self.rev_units = []  # Initialize an empty list for the reversible
        # units
        self.poles = {}  # Initialize an empty dictionary for the poles

        # An assembly unit is created with at least a production unit and a
        # consumption unit,or a reversible unit.
        # If a reversible unit is added, possibility to add (or not)
        # production and/or consumption units.
        if rev_units:
            if not isinstance(rev_units, list):
                raise TypeError('rev_units should be a list.')
            else:
                # if list or rev_units, adding rev_units
                for rev_unit in rev_units:
                    # rev_units should only contain ReversibleUnit objects
                    if not isinstance(rev_unit, ReversibleUnit):
                        raise TypeError(
                            'The elements in rev_units have to be the'
                            ' type "ReversibleUnit".')
                    else:
                        self._add_reversible_unit(rev_unit)

                # if rev_units is not None, possibility to add prod_units
                # (or not)
                if prod_units is None:
                    pass
                elif not isinstance(prod_units, list):
                    raise TypeError('prod_units should be a list')
                else:
                    for prod_unit in prod_units:
                        # prod_units should only contain ProductionUnit objects
                        if not isinstance(prod_unit, ProductionUnit):
                            raise TypeError(
                                'The elements in prod_units have to be the'
                                ' type "ProductionUnit".')
                        else:
                            self._add_production_unit(prod_unit)

                # if rev_units is not None, possibility to add cons_units
                # (or not)
                if cons_units is None:
                    pass
                elif not isinstance(cons_units, list):
                    raise TypeError('cons_units should be a list')
                else:
                    for cons_unit in cons_units:
                        # cons_units should only contain ConsumptionUnit
                        # objects
                        if not isinstance(cons_unit, ConsumptionUnit):
                            raise TypeError(
                                'The elements in cons_units have to be the'
                                ' type "ConsumptionUnit".')
                        else:
                            self._add_consumption_unit(cons_unit)

        # If there is no reversible unit, the assembly unit needs at least
        # one consumption and one production unit.
        else:
            if not prod_units:
                raise IndexError(
                    'You have to fill at least a production unit.')
            elif not isinstance(prod_units, list):
                raise TypeError('prod_units should be a list.')
            else:
                for prod_unit in prod_units:
                    # prod_units should only contain ProductionUnit objects
                    if not isinstance(prod_unit, ProductionUnit):
                        raise TypeError(
                            'The elements in prod_units have to be the'
                            ' type "ProductionUnit".')
                    else:
                        self._add_production_unit(prod_unit)

            if not cons_units:
                raise IndexError(
                    'You have to fill at least a consumption unit.')
            elif not isinstance(cons_units, list):
                raise TypeError('cons_units should be a list.')
            else:
                for cons_unit in cons_units:
                    # cons_units should only contain ConsumptionUnit
                    if not isinstance(cons_unit, ConsumptionUnit):
                        raise TypeError(
                            'The elements in cons_units have to be the'
                            ' type "ConsumptionUnit".')
                    else:
                        self._add_consumption_unit(cons_unit)

    def _add_production_unit(self, prod_unit):
        """
        :param prod_unit: production unit to be added to the
            production_units list
        """
        if prod_unit not in self.prod_units:
            poles_nb = len(self.poles)
            self.poles[poles_nb + 1] = prod_unit.poles[1]
            self.prod_units.append(prod_unit)
            prod_unit.parent = self
        else:
            print('Production unit {0} already in the production_units '
                  'list'.format(prod_unit.name))

    def _add_consumption_unit(self, cons_unit):
        """
        :param cons_unit: consumption unit to be added to the
            consumption_units list
        """
        if cons_unit not in self.cons_units:
            poles_nb = len(self.poles)
            self.poles[poles_nb + 1] = cons_unit.poles[1]
            self.cons_units.append(cons_unit)
            cons_unit.parent = self
        else:
            print('Consumption unit {0} already in the consumption_units '
                  'list'.format(cons_unit.name))

    def _add_reversible_unit(self, rev_unit):
        """
        :param rev_unit: reversible unit to be added to the
            reversible_units list
        """
        if rev_unit not in self.rev_units:
            poles_nb = len(self.poles)
            # Adding the various poles of the reversible_unit to the
            # assembly_unit.
            for p in range(1, len(rev_unit.poles) + 1):
                self.poles[poles_nb + p] = rev_unit.poles[p]
            self.rev_units.append(rev_unit)
            rev_unit.parent = self
        else:
            print('Reversible unit {0} already in the reversible_units list'
                  .format(rev_unit.name))

    def minimize_exergy_destruction(self, weight=1):
        """ This is the main objective of any exergetic optimization. """
        if hasattr(self, 'exergy_dest'):
            self.min_exergy_dest = Objective(name='min_exergy_destruction',
                                             exp='lpSum({0}_exergy_dest[t] '
                                                 'for t in time.I)'.
                                             format(self.name), weight=weight,
                                             parent=self)
        else:
            raise ValueError("You should initialize exergy calculations "
                             "on the unit named {0} before trying to "
                             "minimize its exergy destruction."
                             .format(self.name))



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module enables to model buildings as thermal loads**

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import numpy as np
import pandas as pd
from pulp import LpContinuous, LpBinary

from ...energy.units.consumption_units import VariableConsumptionUnit
from ...general.optimisation.elements import *
from ...general.optimisation.core import OptObject
from ...general.utils.maths import def_abs_value

__docformat__ = "restructuredtext en"

# Stefan-boltzmann constant
BOLTZMANN = 0.000000056697   # W/m2K4

# Thermal resistance of external surfaces according to ISO 6946
RSE = 0.04


class RCNetwork_1(OptObject):
    #                   H_EA
    #                   ____
    # theta_ea     o---|____|----o    T_int
    #                            |
    #                            _
    #                           | |   H_AC
    #                           |_|
    #                   H_EC     |
    #                   ____     |
    # theta_ec     o---|____|----o    theta_c
    #                            |
    #                            _
    #                           | |   H_MC
    #                           |_|
    #                   H_EM     |
    #                   ____     |
    # theta_em     o---|____|----o    theta_m
    #                            |
    #                           ___
    #                           ___   C_M
    #                            |

    def __init__(self, time, name, T_ext, theta_ec, theta_em,
                 T_int_min=0, T_int_max=50, theta_ea=None, theta_c=None,
                 theta_m=None, h_ea=0, h_ac=0, h_ec=0, h_mc=0, h_em=0, c_m=0,
                 f_im=None, f_r_l=0.7, f_r_p=0.5, f_r_a=0.2, f_sa=0.1,
                 f_sm=None, f_sc=None, f_ic=None, f_hc_cv=None, U_wall=0.2,
                 U_win=1.2, U_roof=0.2, e_wall=0.9, e_win=0.9, e_roof=0.9,
                 a_wall=0.6, a_roof=0.6, A_wall=None, A_win=None, A_roof=None,
                 owner=None):

        OptObject.__init__(self, name, 'Thermal_zone')

        self.time = time
        self.owner = owner

        self.F_DICT = {'F_IM': f_im, 'F_R_L': f_r_l, 'F_R_P': f_r_p,
                       'F_R_A': f_r_a, 'F_SA': f_sa, 'F_SM': f_sm,
                       'F_SC': f_sc, 'F_IC': f_ic, 'F_HC_CV': f_hc_cv}

        self.U_DICT = {'WIN': U_win, 'WALL': U_wall, 'ROOF': U_roof}
        self.E_DICT = {'WIN': e_win, 'WALL': e_wall, 'ROOF': e_roof}
        self.A_DICT = {'WIN': A_win, 'WALL': A_wall, 'ROOF': A_roof}
        self.AB_DICT = {'WALL': a_wall, 'ROOF': a_roof}

        # Calculation simplifications
        # TODO : Check
        try:
            h_1 = 1 / (1 / h_ea + 1 / h_ac)
        except ZeroDivisionError:
            h_1 = 0
            # if h_ea == 0:
            #     h_1 = h_ac
            # else:
            #     h_1 = h_ea

        h_2 = h_1 + h_ec
        h_3 = 1 / (1 / h_2 + 1 / h_mc)

        # Cm: Explanation of /3600 (Wh/K) SIA 2044 unit is Wh/K, ISO unit is J/K

        self.H_C_DICT = {'H_EA': h_ea, 'H_AC': h_ac, 'H_EC': h_ec,
                         'H_MC': h_mc, 'H_EM': h_em, 'H_1': h_1, 'H_2': h_2,
                         'H_3': h_3, 'C_M': c_m / 3600}

        # Creation of all nodes' temperatures
        self.T_int = Quantity(name='T_int', lb=T_int_min,
                              ub=T_int_max, vlen=time.LEN,
                              vtype=LpContinuous, parent=self)

        self.T_ext = Quantity(name='T_ext', value=T_ext, vlen=time.LEN,
                              vtype=LpContinuous, parent=self)

        self.theta_ea = Quantity(name='theta_ea', value=theta_ea, lb=-50,
                                 ub=50, vlen=time.LEN, vtype=LpContinuous,
                                 parent=self)

        self.theta_ec = Quantity(name='theta_ec', value=theta_ec,
                                 vlen=time.LEN, vtype=LpContinuous,
                                 parent=self)

        self.theta_c = Quantity(name='theta_c', value=theta_c, lb=-50,
                                ub=50, vlen=time.LEN, vtype=LpContinuous,
                                parent=self)

        self.theta_em = Quantity(name='theta_em', value=theta_em,
                                 vlen=time.LEN, vtype=LpContinuous,
                                 parent=self)

        self.theta_m = Quantity(name='theta_m', value=theta_m, lb=-50,
                                ub=50, vlen=time.LEN, vtype=LpContinuous,
                                parent=self)

        self.T_op = Quantity(name='T_op', lb=-50, ub=50, vlen=time.LEN,
                             vtype=LpContinuous, parent=self)


class ZEA_RCNetwork_1(RCNetwork_1):
    def __init__(self, time, name, T_ext, A_f, A_win, Aext_v, A_roof,
                 footprint, U_win, U_wall, U_roof, U_base, floors, e_wall=0.9,
                 e_win=0.9, e_roof=0.9, a_wall=0.6, a_roof=0.6,
                 construction='heavy', height_bg=0, perimeter=0, f_hc_cv=1,
                 void=0, hvac_prop=None, T_int_min=0, T_int_max=50, owner=None):

        """
            Copyright 2015, Architecture and Building Systems - ETH Zurich

        :param time:
        :param name:
        :param T_ext:
        :param A_f:
        :param A_win:
        :param Aext_v:
        :param A_roof:
        :param footprint:
        :param U_win:
        :param U_wall:
        :param U_roof:
        :param U_base:
        :param floors:
        :param construction:
        :param height_bg:
        :param perimeter:
        :param f_hc_cv:
        :param void:
        :param hvac_prop:
        :param T_int_min:
        :param T_int_max:
        :param owner:
        """

        # Import functions from City Energy Analyst - ETH Zurich
        from cea.demand.rc_model_SIA import calc_h_ea, calc_h_ac, calc_h_ec, \
            calc_h_mc, calc_h_em, \
            calc_h_op_m, calc_f_im, calc_f_sm, calc_f_sc, calc_f_ic, f_r_a, \
            f_r_l, f_r_p, f_sa

        if hvac_prop is not None:
            m_v_sys = hvac_prop.vent_param['M_VE_MECH']
            m_v_w = hvac_prop.vent_param['M_VE_W']
            m_v_inf = hvac_prop.vent_param['M_VE_INF']
        else:
            m_v_sys = 0
            m_v_w = 0
            m_v_inf = 1.23 * 0.1 ** (2/3) * A_f * 0.000277778 * 3
            Warning("You didn't enter any HVAC property for the thermal"
                    " zone properties {}.".format(name))

        # Calculation of the fraction of windows on walls
        window_to_wall_ratio = A_win / Aext_v

        Aop_sup = calc_Aop_sup(Awall_all=Aext_v, void=void,
                               window_to_wall_ratio=window_to_wall_ratio)
        Aop_bel = calc_Aop_bel(height_bg=height_bg, perimeter=perimeter,
                               footprint=footprint)

        A_tot = A_win + Aop_sup + footprint + Aop_bel + A_roof * (floors - 1)

        Htr_op = calc_Htr_op(Aop_bel=Aop_bel, Aop_sup=Aop_sup,
                             footprint=footprint, U_base=U_base, U_wall=U_wall,
                             U_roof=U_roof)
        Htr_w = A_win * U_win

        Cm_Af = get_Cm_Af(construction)

        # Af: conditioned floor area (heated/cooled) in [m2]
        A_m = calc_Am(Cm_Af=Cm_Af, Af=A_f)
        C_m = calc_cm(Cm_Af=Cm_Af, Af=A_f)

        h_ea = calc_h_ea(m_v_sys, m_v_w, m_v_inf)
        h_ac = calc_h_ac(A_tot)
        h_ec = calc_h_ec(Htr_w)
        h_mc = calc_h_mc(A_m)
        h_op_m = calc_h_op_m(Htr_op)
        h_em = calc_h_em(h_op_m, h_mc)

        f_im = calc_f_im(A_tot, A_m)
        f_sm = calc_f_sm(A_tot, A_m, A_win)
        f_sc = calc_f_sc(A_tot, A_m, A_win, h_ec)
        f_ic = calc_f_ic(A_tot, A_m, h_ec)

        RCNetwork_1.__init__(self, time, name, T_ext=T_ext, theta_ec=T_ext,
                             theta_em=T_ext, T_int_min=T_int_min,
                             T_int_max=T_int_max, theta_ea=T_ext,
                             theta_c=None, theta_m=None, h_ea=h_ea,
                             h_ac=h_ac, h_ec=h_ec, h_mc=h_mc, h_em=h_em,
                             c_m=C_m, f_im=f_im, f_r_l=f_r_l, f_r_p=f_r_p,
                             f_r_a=f_r_a, f_sa=f_sa, f_sm=f_sm, f_sc=f_sc,
                             f_ic=f_ic, f_hc_cv=f_hc_cv, U_wall=U_wall,
                             U_win=U_win, U_roof=U_roof, e_wall=e_wall,
                             e_win=e_win, e_roof=e_roof, a_wall=a_wall,
                             a_roof=a_roof, A_wall=Aop_sup, A_win=A_win,
                             A_roof=A_roof, owner=owner)


class ThermalZone(OptObject):

    def __init__(self, rc_network, phi_i_a, phi_i_l, phi_i_p, I_sol_av, Fsh_win,
                 T_mean, T_dew, sky_cover=1, T_ext=None, hvac_prop=None):
        OptObject.__init__(self, rc_network.name, 'Thermal zone with power flows')

        self.time = rc_network.time
        self.owner = rc_network.owner
        self.prop = rc_network
        rc_network.parent = self

        # ++ Creation of energy flows ++
        self.phi_a = Quantity(name='phi_a', vlen=self.time.LEN,
                              vtype=LpContinuous, parent=self)

        self.phi_c = Quantity(name='phi_c', vlen=self.time.LEN,
                              vtype=LpContinuous, parent=self)

        self.phi_m = Quantity(name='phi_m', vlen=self.time.LEN,
                              vtype=LpContinuous, parent=self)

        self.phi_m_tot = Quantity(name='phi_m_tot', vlen=self.time.LEN,
                                  vtype=LpContinuous, parent=self)

        # Internal gains
        self.phi_i_l = Quantity(name='phi_i_l', value=phi_i_l,
                                vlen=self.time.LEN, vtype=LpContinuous,
                                parent=self)

        self.phi_i_p = Quantity(name='phi_i_p', value=phi_i_p,
                                vlen=self.time.LEN, vtype=LpContinuous,
                                parent=self)

        self.phi_i_a = Quantity(name='phi_i_a', value=phi_i_a,
                                vlen=self.time.LEN, vtype=LpContinuous,
                                parent=self)

        # Solar gains
        I_sol = calc_I_sol(I_sol_average=I_sol_av,
                           Aop_sup=self.prop.A_DICT['WALL'],
                           Awin=self.prop.A_DICT['WIN'],
                           Aroof=self.prop.A_DICT['ROOF'],
                           a_wall=self.prop.AB_DICT['WALL'],
                           a_roof=self.prop.AB_DICT['ROOF'],
                           U_wall=self.prop.U_DICT['WALL'],
                           U_roof=self.prop.U_DICT['ROOF'], Fsh_win=Fsh_win)

        self.I_sol = Quantity(name='I_sol', value=I_sol,
                              vlen=self.time.LEN, vtype=LpContinuous,
                              parent=self, unit='kW')

        # Re-irradiated flows
        self.I_rad = Quantity(name='I_rad', vlen=self.time.LEN,
                              vtype=LpContinuous, parent=self)

        # External gains
        self.phi_s = Quantity(name='phi_s', vlen=self.time.LEN,
                              vtype=LpContinuous, parent=self)

        # Heating and cooling
        self.phi_hc_cv = Quantity(name='phi_hc_cv', vlen=self.time.LEN,
                                  vtype=LpContinuous, parent=self)
        self.phi_hc_r = Quantity(name='phi_hc_r', vlen=self.time.LEN,
                                 vtype=LpContinuous, parent=self)
        self.phi_hc = Quantity(name='phi_hc', lb=-10e+8, ub=10e+8,
                               vlen=self.time.LEN, vtype=LpContinuous,
                               parent=self)

        # Get all coefficients from thermal zone properties
        f_im = rc_network.F_DICT['F_IM']
        f_r_l = rc_network.F_DICT['F_R_L']
        f_r_p = rc_network.F_DICT['F_R_P']
        f_r_a = rc_network.F_DICT['F_R_A']
        f_sa = rc_network.F_DICT['F_SA']
        f_sm = rc_network.F_DICT['F_SM']
        f_ic = rc_network.F_DICT['F_IC']
        f_sc = rc_network.F_DICT['F_SC']
        f_hc_cv = rc_network.F_DICT['F_HC_CV']

        h_em = rc_network.H_C_DICT['H_EM']
        h_ea = rc_network.H_C_DICT['H_EA']
        h_ac = rc_network.H_C_DICT['H_AC']
        h_ec = rc_network.H_C_DICT['H_EC']
        h_mc = rc_network.H_C_DICT['H_MC']
        h_1 = rc_network.H_C_DICT['H_1']
        h_2 = rc_network.H_C_DICT['H_2']
        h_3 = rc_network.H_C_DICT['H_3']
        c_m = rc_network.H_C_DICT['C_M']

        # Get ventilation data from hvac_prop
        if hvac_prop is not None:
            m_v_sys = hvac_prop.vent_param['M_VE_MECH']
            m_v_w = hvac_prop.vent_param['M_VE_W']
            m_v_inf = hvac_prop.vent_param['M_VE_INF']
            theta_v_sys = hvac_prop.vent_param['THETA_V_SYS']
        else:
            m_v_sys = 0
            m_v_w = 0
            m_v_inf = 0
            theta_v_sys = None
            Warning("You didn't enter any HVAC property for the thermal"
                    " zone properties {}.".format(rc_network.name))

        # Set all equations into constraints
        self.__set_thermal_flows_equations(f_im, f_r_l, f_r_p, f_r_a, f_sa,
                                           f_sm, f_ic, f_sc, f_hc_cv, h_em,
                                           h_3, h_ec, h_1, h_ea, h_2, T_dew,
                                           T_mean, sky_cover)

        self.__set_temperature_flows_equations(h_ac, h_ea, h_mc, h_ec, h_1, h_3,
                                               h_em, c_m, m_v_sys, theta_v_sys,
                                               m_v_w, m_v_inf)

    def __set_thermal_flows_equations(self, f_im, f_r_l, f_r_p, f_r_a, f_sa,
                                      f_sm, f_ic, f_sc, f_hc_cv, h_em, h_3,
                                      h_ec, h_1, h_ea, h_2, Tdew, Tmean,
                                      sky_cover):

        # Calculation of energy flows
        self.calc_phi_m_tot = DefinitionDynamicConstraint(
            name='calc_phi_m_tot',
            exp_t='{0}_phi_m_tot[t] == {0}_phi_m[t] '
                  '+ {h_em} * {0}_theta_em[t] '
                  '+ ({h_3} * ({0}_phi_c[t] + {h_ec} * {0}_theta_ec[t]'
                  '            + {h_1} * ({0}_phi_a[t] * 1 / {h_ea} '
                  '                        + {0}_theta_ea[t]))) / {h_2}'
                                                .format(self.name, h_em=h_em,
                                                        h_3=h_3,
                                                        h_ec=h_ec, h_1=h_1,
                                                        h_ea=h_ea, h_2=h_2))

        self.calc_phi_m = DefinitionDynamicConstraint(name='calc_phi_m',
                                                      exp_t='{0}_phi_m[t] == '
                                                            '{f_im} * ('
                                                  '{f_r_l} * '
                                                  '{0}_phi_i_l[t] + {f_r_p} * '
                                                  '{0}_phi_i_p[t] + '
                                                  '{f_r_a} * {0}_phi_i_a[t] + '
                                                  '{0}_phi_hc_r[t]) '
                                                  '+ (1 - {f_sa}) * {f_sm} * '
                                                  '{0}_phi_s[t]'
                                                      .format(self.name,
                                                              f_im=f_im,
                                                              f_r_l=f_r_l,
                                                              f_r_p=f_r_p,
                                                              f_r_a=f_r_a,
                                                              f_sa=f_sa,
                                                              f_sm=f_sm))

        self.calc_phi_a = DefinitionDynamicConstraint(name='calc_phi_a',
                                                      exp_t='{0}_phi_a[t] == {f_sa} * {'
                                                  '0}_phi_s[t] + '
                                                  '(1 - {f_r_l}) * {'
                                                  '0}_phi_i_l[t] + (1 - '
                                                  '{f_r_p}) * {0}_phi_i_p[t] '
                                                  '+(1 - {f_r_a}) * '
                                                  '{0}_phi_i_a[t] + {'
                                                  '0}_phi_hc_cv[t]'
                                                      .format(self.name,
                                                              f_sa=f_sa,
                                                              f_r_l=f_r_l,
                                                              f_r_p=f_r_p, f_r_a=f_r_a))

        self.calc_phi_c = DefinitionDynamicConstraint(name='calc_phi_c',
                                                      exp_t='{0}_phi_c[t] == {f_ic} * ('
                                                  '{f_r_l} * '
                                                  '{0}_phi_i_l[t] + {f_r_p} * '
                                                  '{0}_phi_i_p[t] + '
                                                  '{f_r_a} * {0}_phi_i_a[t] + '
                                                  '{0}_phi_hc_r[t]) '
                                                  '+ (1 - {f_sa}) * {f_sc} * '
                                                  '{0}_phi_s[t]'
                                                      .format(self.name,
                                                              f_ic=f_ic,
                                                              f_r_l=f_r_l,
                                                              f_r_p=f_r_p, f_r_a=f_r_a,
                                                              f_sa=f_sa,
                                                              f_sc=f_sc))

        self.calc_phi_s = DefinitionDynamicConstraint(name='calc_phi_s',
                                                      exp_t='{0}_phi_s[t] == {0}_I_sol['
                                                  't] - {0}_I_rad[t]'.format(
                                                self.name))

        # Linearization of the irradiated heat rate flow to the sky
        exp_t = write_linerazation_exp(T_dry=self.prop.T_ext.value,
                                       T_dew=Tdew,
                                       Tlin=Tmean, sky_cover=sky_cover,
                                       U_wall=self.prop.U_DICT['WALL'],
                                       U_win=self.prop.U_DICT['WIN'],
                                       U_roof=self.prop.U_DICT['ROOF'],
                                       e_wall=self.prop.E_DICT['WALL'],
                                       e_win=self.prop.E_DICT['WIN'],
                                       e_roof=self.prop.E_DICT['ROOF'],
                                       A_wall=self.prop.A_DICT['WALL'],
                                       A_win=self.prop.A_DICT['WIN'],
                                       A_roof=self.prop.A_DICT['ROOF'],
                                       name=self.name)

        self.calc_I_rad = DefinitionDynamicConstraint(name='calc_I_rad',
                                                      t_range='for t in time.I[72:]',
                                                      exp_t=exp_t)

        self.calc_phi_hc_cv = DefinitionDynamicConstraint(
            name='calc_phi_hc_cv',
            t_range='for t in time.I',
            exp_t='{0}_phi_hc_cv[t] == {'
                                                      'f_hc_cv} * '
                                                      '{0}_phi_hc[t]'.format(
                                                    self.name,
                                                    f_hc_cv=f_hc_cv))

        self.calc_phi_hc_r = DefinitionDynamicConstraint(name='calc_phi_hc_r',
                                                         t_range='for t in time.I',
                                                         exp_t='{0}_phi_hc_r[t] == (1 '
                                                     '-{f_hc_cv}) * '
                                                     '{0}_phi_hc[t]'.format(
                                                   self.name,
                                                   f_hc_cv=f_hc_cv))

    def __set_temperature_flows_equations(self, h_ac, h_ea, h_mc, h_ec, h_1,
                                          h_3, h_em, c_m, m_v_sys, theta_v_sys,
                                          m_v_w, m_v_inf):
        # Calculation of nodes' temperatures
        self.calc_T_int = DefinitionDynamicConstraint(name='calc_T_int',
                                                      t_range='for t in time.I[72:]',
                                                      exp_t='{0}_T_int[t] == '
                                                            '({h_ac} * '
                                                  '{0}_theta_c[t] + '
                                                  '{h_ea} * {0}_theta_ea[t] + '
                                                  '{0}_phi_a[t]) '
                                                  '/ ({h_ac} + {h_ea})'
                                                      .format(self.name,
                                                              h_ac=h_ac,
                                                              h_ea=h_ea))

        self.calc_theta_c = DefinitionDynamicConstraint(name='calc_theta_c',
                                                        t_range='for t in time.I[72:]',
                                                        exp_t='{0}_theta_c[t] == ({'
                                                    'h_mc} * '
                                                    '({0}_theta_m[t] + {'
                                                    '0}_theta_m[t-1])/2 + '
                                                    '{h_ec} * {0}_theta_ec[t] '
                                                    '+ {0}_phi_c[t] +'
                                                    ' {h_1} * (1/{h_ea} * {'
                                                    '0}_phi_a[t] + '
                                                    '{0}_theta_ea[t])) / ({'
                                                    'h_mc} + {h_ec} + '
                                                    '{h_1})'
                                                        .format(self.name,
                                                                h_mc=h_mc,
                                                                h_ec=h_ec, h_1=h_1,
                                                                h_ea=h_ea))

        self.calc_theta_m = DefinitionDynamicConstraint(name='calc_theta_m',
                                                        t_range='for t in time.I[72:]',
                                                        exp_t='{0}_theta_m['
                                                              't] == ({'
                                                    '0}_theta_m[t-1] * '
                                                    '({c_m} - 0.5 * ({h_3} + '
                                                    '{h_em})) + '
                                                    '{0}_phi_m_tot[t]) / ({'
                                                    'c_m} + 0.5 * '
                                                    '({h_3} + {h_em}))'
                                                        .format(self.name,
                                                                c_m=c_m,
                                                                h_em=h_em,
                                                                h_3=h_3))

        self.calc_theta_m_0 = DefinitionConstraint(name='calc_theta_m_0',
                                                   exp='{0}_theta_m[0] == {0}_T_ext[0]'
                                                   .format(self.name), parent=self)

        # Time for dynamics to appear
        # self.calc_operative_temp = DefinitionDynamicConstraint(
        #     t_range='for t in time.I[144:]',
        #     name='calc_operative_temp', exp_t='2 * {0}_T_op[t] == '
        #                                       '{0}_theta_m[t] + {0}_T_int[t]'
        #         .format(self.name), parent=self)
        self.calc_operative_temp = DefinitionDynamicConstraint(
            t_range='for t in time.I[144:]',
            name='calc_operative_temp',
            exp_t='100 * {0}_T_op[t] == 31 * {0}_T_int[t] + 69 * {'
                  '0}_theta_c[t]'
                .format(self.name), parent=self)

        if m_v_sys != 0 and theta_v_sys is not None:
            # Intermediary temperature to calculate theta_ea in case of
            # mechanical ventilation
            self.theta_v_sys = Quantity(name='theta_v_sys', value=theta_v_sys,
                                        lb=0, ub=50,
                                        vlen=self.time.LEN, vtype=LpContinuous,
                                        parent=self)

            self.calc_theta_ea = DefinitionDynamicConstraint(
                name='calc_theta_ea', t_range='for t in time.I[72:]',
                exp_t='{0}_theta_ea[t] == ({m_v_sys} * {0}_theta_v_sys[t] '
                      '+ ({m_v_w} + {m_v_inf}) * {0}_T_ext[t]) '
                      '* 1 / ({m_v_sys} + {m_v_w} + {m_v_inf})'
                                                   .format(self.name,
                                                           m_v_sys=m_v_sys,
                                                           m_v_w=m_v_w,
                                                           m_v_inf=m_v_inf))

        else:
            # If no mechanical ventilation, theta_ea equals the external
            # temperature
            self.calc_theta_ea = DefinitionDynamicConstraint(
                name='calc_theta_ea',
                t_range='for t in time.I',
                exp_t='{0}_theta_ea[t] == '
                                                         '{0}_T_ext[t]'
                .format(self.name))

    def split_heating_and_cooling(self, p_max_heating=10e+12,
                                  p_max_cooling=10e+12):
        self.phi_heating = Quantity(name='phi_heating', lb=0, ub=p_max_heating,
                                    vlen=self.time.LEN,
                                    vtype=LpContinuous, parent=self)

        self.phi_cooling = Quantity(name='phi_cooling', lb=0, ub=p_max_cooling,
                                    vlen=self.time.LEN,
                                    vtype=LpContinuous, parent=self)

        self.heating_on = Quantity(name='heating_on', vlen=self.time.LEN,
                                   vtype=LpBinary, parent=self)

        self.split_heat_cool = DefinitionDynamicConstraint(
            exp_t='{0}_phi_hc[t] == {0}_phi_heating[t] - {0}_phi_cooling[t]'
                .format(self.name), t_range='for t in time.I',
            name='split_heat_cool', parent=self)

        self.heating_only = DefinitionDynamicConstraint(
            exp_t='{0}_phi_heating[t] <= {0}_heating_on[t] * {p_max_h}'
                .format(self.name, p_max_h=p_max_heating), name='heating_only')

        self.cooling_only = DefinitionDynamicConstraint(
            exp_t='{0}_phi_cooling[t] <= (1 - {0}_heating_on[t]) * {p_max_c}'
                .format(self.name, p_max_c=p_max_cooling), name='cooling_only')


class HeatingLoad(VariableConsumptionUnit):
    """

    """

    def __init__(self, time, name, tz, p_max=10e+12, T_set=19, temp_margin=1,
                 no_cooling=True):
        self.tz = tz
        self.T_set = T_set
        self.temp_margin = temp_margin
        tz.parent = self

        if not isinstance(tz, ThermalZone):
            raise TypeError("tz should be a ThermalZone !")

        print('Redefining minimal and maximal air temperatures for '
              'thermal zone {0}'.format(tz.name))
        self.tz.prop.T_op.lb = T_set - temp_margin
        self.tz.prop.T_op.ub = T_set + temp_margin

        if no_cooling:
            p_max_cooling = 1e5
        else:
            p_max_cooling = 1e9
        self.tz.split_heating_and_cooling(p_max_heating=1e9,   # p_max
                                          p_max_cooling=p_max_cooling)   # kW

        VariableConsumptionUnit.__init__(self, time, name, energy_type='Heat',
                                         operator=tz.owner, pmax=p_max,
                                         e_max=10e9)
        self.calc_e_tot = None
        self.e_tot = None

        self.def_heat_cons = DefinitionDynamicConstraint(
            exp_t='{0}_p[t] * 1000 == {1}_phi_heating[t]'   # kW / 1000
                .format(self.name, self.tz.name), name='def_heat_cons',
            t_range='for t in time.I[144:]', parent=self)

    def maximize_thermal_comfort(self, T_op=None, weight=1):
        """

        :param T_op: Operative temperature wished for the maximal thermal
            comfort
        :param weight: Weight of the objective
        """
        if T_op is None:
            T_op = self.T_set

        self.diff_Top_opt = Quantity(name='diff_Top_opt', opt=True,
                                     lb=-self.temp_margin,
                                     ub=self.temp_margin,
                                     vlen=self.time.LEN, parent=self)

        if isinstance(T_op, (int, float)):
            self.def_diff_Top_opt = DefinitionDynamicConstraint(
                name='def_diff_Top_opt',
                exp_t='{0}_diff_Top_opt[t] == {1}_T_op[t] - {2}'.format(
                    self.name, self.tz.prop.name, T_op), parent=self)
        elif isinstance(T_op, list):
            self.def_diff_Top_opt = DefinitionDynamicConstraint(
                name='def_diff_Top_opt',
                exp_t='{0}_diff_Top_opt[t] == {1}_T_op[t] - {1}[t]'.format(
                    self.name, self.tz.prop.name, T_op), parent=self)

        diff = def_abs_value(self.diff_Top_opt, q_max=self.temp_margin,
                             q_min=-self.temp_margin)

        self.max_th_comfort = Objective(name='max_th_comfort',
                                        exp='lpSum({0}_{1}[t] for t in '
                                             'time.I[72:])'.format(self.name,
                                                              diff.name),
                                        parent=self)


def calc_Am(Cm_Af, Af):
    # Am: Effective mass area in [m2]
    Am = lookup_effective_mass_area_factor(Cm_Af) * Af
    return Am


def calc_Aop_bel(height_bg, perimeter, footprint):
    # Aop_bel: Opaque areas below ground (including ground floor,
    # excluding voids and windows) [m2]
    Aop_bel = height_bg * perimeter + footprint
    return Aop_bel


def calc_Aop_sup(Awall_all, void, window_to_wall_ratio):
    # Aop_sup: Opaque wall areas above ground (excluding voids
    # and windows) [m2]
    Aop_sup = Awall_all * (1 - void) * (1 - window_to_wall_ratio)
    return Aop_sup


def calc_cm(Cm_Af, Af):
    # Cm: internal heat capacity in [J/k]
    Cm = Cm_Af * Af
    return Cm


def calc_Hg(Aop_bel, U_base):
    # Hg: steady-state thermal transmission coefficient to the ground
    # in [W/K]
    Hg = 0.7 * Aop_bel * U_base

    return Hg


def calc_Hd(Aop_sup, U_wall, footprint, U_roof):
    # HD: direct thermal transmission  coefficient to  the
    # external environment in [W/K])

    Hd = Aop_sup * U_wall + footprint * U_roof

    return Hd


def calc_Htr_op(Aop_bel, Aop_sup, footprint, U_base, U_wall, U_roof):
    Hg = calc_Hg(Aop_bel=Aop_bel, U_base=U_base)
    Hd = calc_Hd(Aop_sup=Aop_sup, U_wall=U_wall, footprint=footprint,
                 U_roof=U_roof)
    Htr_op = Hg + Hd

    return Htr_op


def get_Cm_Af(construction):
    """
        Description	       code	    Cm_Af
        Light construction	T1	    110000
        Medium construction	T2	    165000
        Heavy construction	T3	    300000

    ..
        Copyright 2015, Architecture and Building Systems - ETH Zurich
    """
    if construction == 'light':
        cm_af = 110000
    elif construction == 'medium':
        cm_af = 165000
    elif construction == 'heavy':
        cm_af = 300000
    else:
        raise ValueError('The construction should be "light", "medium" or '
                         '"heavy", but equals {}'.format(construction))

    return cm_af


def lookup_effective_mass_area_factor(cm):
    """
        Look up the factor to multiply the conditioned floor area by to get
        the effective mass area by building
        construction type.
        This is used for the calculation of the effective mass area "Am" in
        `get_prop_RC_model`.
        Standard values can be found in the Annex G of ISO EN13790

        :param: cm: The internal heat capacity per unit of area [J/m2].

        :return: Effective mass area factor (0, 2.5 or 3.2 depending on cm
            value).

    ..
        Copyright 2015, Architecture and Building Systems - ETH Zurich
    """

    if cm == 0:
        return 0
    elif 0 < cm <= 165000.0:
        return 2.5
    else:
        return 3.2


def calc_I_sol(I_sol_average, Aop_sup, Aroof, Awin, a_wall, a_roof, U_wall,
               U_roof, Fsh_win):
    """

    :param I_sol_average: W/m2
    :param Aop_sup: Opaque wall areas above ground (excluding voids and
        windows) [m2]
    :param Aroof: Roof area [m2]
    :param Awin: Windows area [m2]
    :param a_wall: Absorption coefficient of the walls [0..1]
    :param a_roof: Absorption coefficeint of the roof [0..1]
    :param U_wall:
    :param U_roof:
    :param Fsh_win: Shading factor for windows
    :return: I_sol [kW]
    """

    from cea.demand.constants import RSE, F_F

    Asol_wall = Aop_sup * a_wall * RSE * U_wall
    Asol_roof = Aroof * a_roof * RSE * U_roof

    if isinstance(I_sol_average, list):
        if isinstance(Fsh_win, list):
            Asol_win = [Fsh_w * Awin * (1 - F_F) for Fsh_w in Fsh_win]
            I_sol = [I_sol_av * (Asol_wall + Asol_roof + Asol_w) for
                     I_sol_av, Asol_w in zip(I_sol_average, Asol_win)]
        elif isinstance(Fsh_win, (int, float)):
            Asol_win = Fsh_win * Awin * (1 - F_F)
            I_sol = [I_sol_av * (Asol_wall + Asol_roof + Asol_win) for
                     I_sol_av in I_sol_average]
        else:
            raise TypeError('Fsh_win should be an int, a float or a list.')

    elif isinstance(I_sol_average, (int, float)):
        if isinstance(Fsh_win, list):
            Asol_win = [Fsh_w * Awin * (1 - F_F) for Fsh_w in Fsh_win]
            I_sol = [I_sol_average * (Asol_wall + Asol_roof + Asol_w)
                     for Asol_w in Asol_win]
        elif isinstance(Fsh_win, (int, float)):
            Asol_win = [Fsh_win * Awin * (1 - F_F)]
            I_sol = I_sol_average * (Asol_wall + Asol_roof + Asol_win)
        else:
            raise TypeError('Fsh_win should be an int, a float or a list.')
    else:
        raise TypeError('I_sol_average should be an int, a float or a list.')

    return I_sol


def calc_T_sky(T_dry, T_dew, sky_cover=1):
    """

    :param T_dry: Dry bulb temperature in Celsius
    :param T_dew: Dew point temperature in Celsius
    :param sky_cover: Sky cover
    """
    sky_T = np.vectorize(calc_skytemp)(T_dry, T_dew, 1)

    return sky_T  # sky temperature in C


def calc_I_rad_linearization_coef(Tdry, Tdew, Tlin, sky_cover=1):
    """

    :param T_dry: Dry bulb temperature in Celsius
    :param T_dew: Dew point temperature in Celsius
    :param sky_cover:
    :return: list(A), list(B):
    """
    temp_df = pd.DataFrame()
    temp_df['T_sky (C)'] = calc_T_sky(T_dry=Tdry, T_dew=Tdew,
                                      sky_cover=sky_cover)
    temp_df['T_sky (K)'] = temp_df['T_sky (C)'] + 273
    temp_df['T_lin (C)'] = Tlin
    temp_df['T_lin (K)'] = temp_df['T_lin (C)'] + 273

    A = 3 * temp_df['T_lin (K)'] ** 4 + temp_df['T_sky (K)'] ** 4 + 4 * \
        temp_df['T_sky (K)'] * temp_df['T_lin (K)'] ** 3

    B = - 4 * temp_df['T_lin (K)'] ** 3 - 6 * temp_df['T_sky (K)'] * temp_df[
        'T_lin (K)'] ** 2 + 2 * temp_df['T_sky (K)'] ** 3

    return list(A), list(B)


def write_linerazation_exp(T_dry, T_dew, sky_cover, Tlin, U_win, U_wall,
                           U_roof, e_win, e_wall, e_roof, A_win, A_wall,
                           A_roof, name):
    A, B = calc_I_rad_linearization_coef(Tdry=T_dry, Tdew=T_dew, Tlin=Tlin,
                                         sky_cover=sky_cover)

    alpha = RSE * BOLTZMANN / 2

    win_coef = 0.5 * U_win * A_win * e_win
    wall_coef = 0.5 * U_wall * A_wall * e_wall
    roof_coef = U_roof * A_roof * e_roof

    coef_tot = alpha * (win_coef + wall_coef + roof_coef)

    A = [a * coef_tot for a in A]
    B = [b * coef_tot for b in B]

    exp_t = '{A}[t] + {B}[t] * ({0}_theta_c[t-1] + 273)'.format(name, A=A, B=B)

    exp_tot = '{0}_I_rad[t] == '.format(name) + exp_t

    return exp_tot


def calc_skytemp(Tdrybulb, Tdewpoint, N=1):
    """

        Copyright 2014, Architecture and Building Systems - ETH Zurich

    :param Tdrybulb: Drybuld temperature [°C]
    :param Tdewpoint: Dewpoint temperature [°C]
    :param N: Sky cover

    :return: Sky temperature in °C

    """

    import math
    sky_e = (0.787 + 0.764 * math.log((Tdewpoint + 273) / 273)) * 1 \
            + 0.0224 * N + 0.0035 * N ** 2 + 0.00025 * N ** 3
    hor_IR = sky_e * BOLTZMANN * (Tdrybulb + 273) ** 4
    sky_T = ((hor_IR / BOLTZMANN) ** 0.25) - 273

    return sky_T  # sky temperature in C




#! usr/bin/env python3
#  coding=utf-8 #

"""
**This module defines the reversible units**

 The reversible_units module defines various kinds of reversible units with
 associated attributes and methods, from simple to specific ones, inheriting
 from AssemblyUnit.

 It includes :
    - ReversibleUnit : simple reversible unit with only one consumption and
    one production units. It can both produce and consume energy but not at the
    same time.

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from pulp import LpBinary

from .energy_units import EnergyUnit, VariableEnergyUnit, FixedEnergyUnit, \
    SeveralEnergyUnit, SquareEnergyUnit, ShiftableEnergyUnit, AssemblyUnit
from .production_units import ProductionUnit, VariableProductionUnit, \
    FixedProductionUnit
from .consumption_units import ConsumptionUnit, FixedConsumptionUnit, \
    VariableConsumptionUnit
from ...general.optimisation.elements import *
from ...general.optimisation.elements import Objective
from ...general.optimisation.elements import Quantity

__docformat__ = "restructuredtext en"


class ReversibleUnit(AssemblyUnit):
    """
    **Description**

        Simple Reversible unit inheriting from AssemblyUnit. It is made of a
        consumption unit and a production unit that can both operate but not
        at the same time (reversible constraint).

    **Attributes**

     * production_unit (ProductionUnit)
     * consumption_unit (ConsumptionUnit)
     * def_rev (DefinitionDynamicConstraint): definition of the reversible
     constraint
     * def_rev_c (DefinitionDynamicConstraint): definition of the reversible
     constraint in the case where only the consumption is fixed
     * def_rev_p (DefinitionDynamicConstraint): definition of the reversible
     constraint in the case where only the production is fixed

    """

    def __init__(self, time, name, pmin_cons=1e-5, pmax_cons=1e+5,
                 p_cons=None, pmin_prod=1e-5, pmax_prod=1e+5, p_prod=None,
                 energy_type_prod=None, energy_type_cons=None, verbose=True):
        """
        :param time: time used in the studies period
        :param name: name of the reversible unit
        :param pmin_cons: minimal power for the consumption of the reversible
        unit [kW]
        :param pmax_cons: maximal power for the consumption of the reversible
        unit [kW]
        :param p_cons: power values for the consumption of the reversible
        unit [kW]
        :param pmin_prod: minimal power for the production of the reversible
        unit [kW]
        :param pmax_prod: maximal power for the production of the reversible
        unit [kW]
        :param p_prod: power values for the production of the reversible
        unit [kW]
        :param energy_type_prod: produced energy type
        :param energy_type_cons: consumed energy type
        :param verbose:

        """

        if p_prod is None:
            self.production_unit = VariableProductionUnit(
                time, name + '_prod', p_min=pmin_prod, p_max=pmax_prod,
                energy_type=energy_type_prod, verbose=verbose)
        else:
            self.production_unit = FixedProductionUnit(
                time, name + '_prod', p=p_prod,
                energy_type=energy_type_prod, verbose=verbose)

        if p_cons is None:
            self.consumption_unit = VariableConsumptionUnit(
                time, name + '_cons', p_min=pmin_cons, p_max=pmax_cons,
                energy_type=energy_type_cons, verbose=verbose)
        else:
            self.consumption_unit = FixedConsumptionUnit(
                time, name + '_cons', p=p_cons, energy_type=energy_type_cons,
                verbose=verbose)

        AssemblyUnit.__init__(self, time, name,
                              prod_units=[self.production_unit],
                              cons_units=[self.consumption_unit],
                              verbose=verbose)

        # CONSTRAINTS
        if p_cons is None and p_prod is None:
            self.def_rev = DefinitionDynamicConstraint(
                exp_t='{0}_p[t] - (1 - {1}_u[t]) * {2} <= 0'.format(
                    self.production_unit.name, self.consumption_unit.name,
                    pmax_prod),
                t_range='for t in time.I',
                name='def_rev', parent=self)

        elif p_prod is None:
            # The consumption unit is a FixedEnergyUnit and no u attribute
            # is defined for it. The reversible constraint uses the u
            # attribute of the production unit.
            self.def_rev_c = DefinitionDynamicConstraint(
                exp_t='{0}_p[t] - (1 - {1}_u[t]) * {2} <= 0'.format(
                    self.consumption_unit.name, self.production_unit.name,
                    pmax_cons),
                t_range='for t in time.I',
                name='def_rev_c', parent=self)

        elif p_cons is None:
            # The production unit is a FixedEnergyUnit and no u attribute
            # is defined for it. The reversible constraint uses the u
            # attribute of the consumption unit.
            self.def_rev_p = DefinitionDynamicConstraint(
                exp_t='{0}_p[t] - (1 - {1}_u[t]) * {2} <= 0'.format(
                    self.production_unit.name, self.consumption_unit.name,
                    pmax_prod),
                t_range='for t in time.I',
                name='def_rev_p', parent=self)

        else:
            # If both production and consumption are fixed, they should be
            # consistent with the reversible unit behaviour.
            for (pc, pp) in zip(self.production_unit.p.get_value(),
                                self.consumption_unit.p.get_value()):
                if pc != 0 and pp != 0:
                    raise TypeError("The reversible unit power values for "
                                    "production and consumption should be "
                                    "consistent: there cannot be consumption "
                                    "and production at the same time step.")
                else:
                    pass



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module defines the conversion units, inheriting from AssemblyUnits**

The conversion_units module defines various classes of conversion units,
from generic to specific ones.

It includes :
 - ConversionUnit : simple conversion unit. It inherits from AssemblyUnit.
 - ElectricalToThermalConversionUnit : Electrical to thermal Conversion unit
   with an electricity consumption and a thermal production linked by and
   electrical to thermal ratio. It inherits from ConversionUnit
 - HeatPump : Simple Heat Pump with an electricity consumption, a heat
   production and a heat consumption. It has a theoretical coefficient of
   performance COP and inherits from ConversionUnit.

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from .energy_units import AssemblyUnit
from .consumption_units import FixedConsumptionUnit, VariableConsumptionUnit
from .production_units import FixedProductionUnit, VariableProductionUnit
from .reversible_units import ReversibleUnit
from ...general.optimisation.elements import *

__docformat__ = "restructuredtext en"


class ConversionUnit(AssemblyUnit):
    """
    **Description**

        Simple Conversion unit, inheriting from AssemblyUnit

    **Attributes**

     * time : TimeUnit describing the studied time period
     * prod_units : list of the production units
     * cons_units : list of the consumption units
     * poles : dictionary of the poles of the conversion unit

    """

    def __init__(self, time, name, prod_units=None, cons_units=None,
                 rev_units=None, verbose=True):
        AssemblyUnit.__init__(self, time=time, name=name,
                              prod_units=prod_units,
                              cons_units=cons_units,
                              rev_units=rev_units, verbose=verbose)


class ElectricalToThermalConversionUnit(ConversionUnit):
    """
    **Description**

        Electrical to thermal Conversion unit with an electricity consumption
        and a thermal production

    **Attributes**

     * thermal_production_unit : thermal production unit (thermal output)
     * elec_consumption_unit : electricity consumption unit (electrical
       input)
     * conversion : Definition Dynamic Constraint linking the electrical
     input to
       the thermal output through the electrical to thermal ratio

    """

    def __init__(self, time, name, pmin_in_elec=1e-5, pmax_in_elec=1e+5,
                 p_in_elec=None, pmin_out_therm=1e-5, pmax_out_therm=1e+5,
                 p_out_therm=None, elec_to_therm_ratio=1,
                 verbose=True):

        """
        :param time: TimeUnit describing the studied time period
        :param name: name of the electrical to thermal conversion unit
        :param pmin_in_elec: minimal incoming electrical power
        :param pmax_in_elec: maximal incoming electrical power
        :param p_in_elec: power input for the electrical consumption unit
        :param pmin_out_therm: minimal power output (thermal)
        :param pmax_out_therm: maximal power output (thermal)
        :param p_out_therm: power output (thermal)
        :param elec_to_therm_ratio: electricity to thermal ratio <=1
        """

        if p_out_therm is None:
            self.thermal_production_unit = VariableProductionUnit(
                time, name + '_therm_prod', energy_type='Thermal',
                p_min=pmin_out_therm, p_max=pmax_out_therm,
                verbose=verbose)
        else:
            self.thermal_production_unit = FixedProductionUnit(
                time, name + '_therm_prod', energy_type='Thermal',
                p=p_out_therm, verbose=verbose)

        if p_in_elec is None:
            self.elec_consumption_unit = VariableConsumptionUnit(
                time, name + '_elec_cons', p_min=pmin_in_elec,
                p_max=pmax_in_elec, energy_type='Electrical',
                verbose=verbose)
        else:
            self.elec_consumption_unit = FixedConsumptionUnit(
                time, name + '_elec_cons', p=p_in_elec,
                energy_type='Electrical', verbose=verbose)

        ConversionUnit.__init__(self, time, name,
                                prod_units=[self.thermal_production_unit],
                                cons_units=[self.elec_consumption_unit])

        if isinstance(elec_to_therm_ratio, (int, float)):  # e2h_ratio is a
            # mean value
            if elec_to_therm_ratio <= 1:
                self.conversion = DefinitionDynamicConstraint(
                    exp_t='{0}_p[t] == {1} * {2}_p[t]'.format(
                        self.thermal_production_unit.name,
                        elec_to_therm_ratio,
                        self.elec_consumption_unit.name),
                    t_range='for t in time.I', name='conversion', parent=self)
            else:
                raise ValueError('The elec_to_therm_ratio should be lower '
                                 'than 1 (therm_production<elec_consumption)')

        elif isinstance(elec_to_therm_ratio, list):  # e2h_ratio is a list of
            # values
            if len(elec_to_therm_ratio) == self.time.LEN:  # it must have the
                #  right size, i.e. the TimeUnit length.
                if all(e <= 1 for e in elec_to_therm_ratio):
                    self.conversion = DefinitionDynamicConstraint(
                        exp_t='{0}_p[t] == {1}[t] * {2}_p[t]'.format(
                            self.thermal_production_unit.name,
                            elec_to_therm_ratio,
                            self.elec_consumption_unit.name),
                        t_range='for t in time.I', name='conversion',
                        parent=self)
                else:
                    raise ValueError(
                        'The elec_to_therm_ratio values should be '
                        'lower than 1 (therm_production<elec_'
                        'consumption)')
            else:
                raise IndexError('The length of the elec_to_therm_ratio '
                                 'vector should be of the same length as the '
                                 'TimeUnit of the studied period')

        elif isinstance(elec_to_therm_ratio, dict):  # e2h_ratio is a dict of
            # values
            if len(elec_to_therm_ratio) == self.time.LEN:
                if all(e <= 1 for e in elec_to_therm_ratio.values()):
                    self.conversion = DefinitionDynamicConstraint(
                        exp_t='{0}_p[t] == {1}[t] * {2}_p[t]'.format(
                            self.thermal_production_unit.name,
                            elec_to_therm_ratio,
                            self.elec_consumption_unit.name),
                        t_range='for t in time.I', name='conversion',
                        parent=self)
                else:
                    raise ValueError(
                        'The elec_to_therm_ratio values should be '
                        'lower than 1 (therm_production<elec_'
                        'consumption)')
            else:
                raise IndexError('The length of the elec_to_therm_ratio '
                                 'dictionary should be of the same length as '
                                 'the TimeUnit of the studied period')
        else:
            raise TypeError(
                "Electricity to thermal ratio should be a mean value or a "
                "vector (list or dict) on the whole time period !")


class ElectricalConversionUnit(ConversionUnit):
    """
    **Description**

        Electrical Conversion unit with an electricity consumption
        and an electricity production

    **Attributes**

     * elec_production_unit : electricity production unit (electrical output)
     * elec_consumption_unit : electricity consumption unit (electrical
       input)
     * conversion : Definition Dynamic Constraint linking the electrical
     input to the electrical output through the elec_to_elec ratio

    """

    def __init__(self, time, name, pmin_in_elec=1e-5, pmax_in_elec=1e+5,
                 p_in_elec=None, pmin_out_elec=1e-5, pmax_out_elec=1e+5,
                 p_out_elec=None, elec_to_elec_ratio=1):
        """
        :param time: TimeUnit describing the studied time period
        :param name: name of the electrical to elec conversion unit
        :param pmin_in_elec: minimal incoming electrical power
        :param pmax_in_elec: maximal incoming electrical power
        :param p_in_elec: power input for the electrical consumption unit
        :param pmin_out_elec: minimal power output (elec)
        :param pmax_out_elec: maximal power output (elec)
        :param p_out_elec: power output (elec)
        :param elec_to_elec_ratio: electricity to elec ratio <=1
        """

        if p_out_elec is None:
            self.elec_production_unit = VariableProductionUnit(
                time, name + '_elec_prod', energy_type='Electrical',
                p_min=pmin_out_elec, p_max=pmax_out_elec)
        else:
            self.elec_production_unit = FixedProductionUnit(
                time, name + '_elec_prod', energy_type='Electrical',
                p=p_out_elec)

        if p_in_elec is None:
            self.elec_consumption_unit = VariableConsumptionUnit(
                time, name + '_elec_cons', p_min=pmin_in_elec,
                p_max=pmax_in_elec, energy_type='Electrical')
        else:
            self.elec_consumption_unit = FixedConsumptionUnit(
                time, name, p=p_in_elec, energy_type='Electrical')

        ConversionUnit.__init__(self, time, name,
                                prod_units=[self.elec_production_unit],
                                cons_units=[self.elec_consumption_unit])

        if isinstance(elec_to_elec_ratio, (int, float)):  # e2h_ratio is a
            # mean value
            if elec_to_elec_ratio <= 1:
                self.conversion = DefinitionDynamicConstraint(
                    exp_t='{0}_p[t] == {1} * {2}_p[t]'.format(
                        self.elec_production_unit.name,
                        elec_to_elec_ratio,
                        self.elec_consumption_unit.name),
                    t_range='for t in time.I', name='conversion', parent=self)
            else:
                raise ValueError('The elec_to_elec_ratio should be lower '
                                 'than 1 (elec_production<elec_consumption)')

        elif isinstance(elec_to_elec_ratio, list):  # e2h_ratio is a list of
            # values
            if len(elec_to_elec_ratio) == self.time.LEN:  # it must have the
                #  right size, i.e. the TimeUnit length.
                if all(e <= 1 for e in elec_to_elec_ratio):
                    self.conversion = DefinitionDynamicConstraint(
                        exp_t='{0}_p[t] == {1}[t] * {2}_p[t]'.format(
                            self.elec_production_unit.name,
                            elec_to_elec_ratio,
                            self.elec_consumption_unit.name),
                        t_range='for t in time.I', name='conversion',
                        parent=self)
                else:
                    raise ValueError('The elec_to_elec_ratio values should be '
                                     'lower than 1 (elec_production<elec_'
                                     'consumption)')
            else:
                raise IndexError('The length of the elec_to_elec_ratio '
                                 'vector should be of the same length as the '
                                 'TimeUnit of the studied period')

        elif isinstance(elec_to_elec_ratio, dict):  # e2h_ratio is a dict of
            # values
            if len(elec_to_elec_ratio) == self.time.LEN:
                if all(e <= 1 for e in elec_to_elec_ratio.values()):
                    self.conversion = DefinitionDynamicConstraint(
                        exp_t='{0}_p[t] == {1}[t] * {2}_p[t]'.format(
                            self.elec_production_unit.name,
                            elec_to_elec_ratio,
                            self.elec_consumption_unit.name),
                        t_range='for t in time.I', name='conversion',
                        parent=self)
                else:
                    raise ValueError('The elec_to_elec_ratio values should be '
                                     'lower than 1 (elec_production<elec_'
                                     'consumption)')
            else:
                raise IndexError('The length of the elec_to_elec_ratio '
                                 'dictionary should be of the same length as '
                                 'the TimeUnit of the studied period')
        else:
            raise TypeError(
                "Electricity to elec ratio should be a mean value or a "
                "vector (list or dict) on the whole time period !")


class ReversibleConversionUnit(ConversionUnit):
    """
    **Description**

        Reversible Conversion unit with two reversible units,
        one for each side of the conversion unit. Theses sides will be
        called upstream and downstream.

    **Attributes**

     * rev_unit_upstream: reversible unit upstream
     * rev_unit_downstream: reversible unit downstream
     * conversion_up2down: Definition Dynamic Constraint linking the
     consumption of the
     upstream reversible unit to the production of the downstream reversible
     unit through the up2down_eff
     * conversion_down2up:  Definition Dynamic Constraint linking the
     consumption of the
     downstream reversible unit to the production of the upstream reversible
     unit through the down2up_eff

    """

    def __init__(self, time, name, pmin_up=1e-5, pmax_up=1e+5, pmin_down=1e-5,
                 pmax_down=1e+5, up2down_eff=1, down2up_eff=1,
                 energy_type_up=None, energy_type_down=None, verbose=True):
        """

        :param time: time for the studied time period
        :param name: name of the reversible conversion unit
        :param pmin_up: pmin for both consumption and production unit of the
        upstream reversible unit
        :param pmax_up: pmax for both consumption and production unit of the
        upstream reversible unit
        :param pmin_down: pmin for both consumption and production unit of the
        downstream reversible unit
        :param pmax_down: pmax for both consumption and production unit of the
        downstream reversible unit
        :param up2down_eff: efficiency of the conversion from upstream to
        downstream
        :param down2up_eff: efficiency of the conversion from downstream to
        upstream
        :param energy_type_up: energy type of the upstream reversible unit
        :param energy_type_down: energy type of the downstream reversible unit
        """

        self.rev_unit_upstream = \
            ReversibleUnit(time=time, name=name + '_upstream',
                           pmin_cons=pmin_up,
                           pmax_cons=pmax_up, pmin_prod=pmin_up,
                           pmax_prod=pmax_up, energy_type_prod=energy_type_up,
                           energy_type_cons=energy_type_up, verbose=verbose)

        self.rev_unit_downstream = \
            ReversibleUnit(time=time, name=name + '_downstream',
                           pmin_cons=pmin_down, pmax_cons=pmax_down,
                           pmin_prod=pmin_down, pmax_prod=pmax_down,
                           energy_type_cons=energy_type_down, verbose=verbose)

        ConversionUnit.__init__(self, time, name,
                                rev_units=[self.rev_unit_upstream,
                                           self.rev_unit_downstream])

        # Managing the upstream to downstream conversion for various types
        # of up2down_eff:

        if isinstance(up2down_eff, (int, float)):  # up2down_eff is a mean
            # value
            if up2down_eff <= 1:
                self.conversion_up2down = DefinitionDynamicConstraint(
                    exp_t='{0}_p[t] == {1} * {2}_p[t]'.format(
                        self.rev_unit_downstream.production_unit.name,
                        up2down_eff,
                        self.rev_unit_upstream.consumption_unit.name),
                    t_range='for t in time.I', name='conversion_up2down',
                    parent=self)
            else:
                raise ValueError('The up2down_efficiency should be lower '
                                 'than 1')

        elif isinstance(up2down_eff, list):  # up2down_eff is a list of
            # values
            if len(up2down_eff) == self.time.LEN:  # it must have the right
                # size, i.e. the TimeUnit length.
                if all(e <= 1 for e in up2down_eff):
                    self.conversion_up2down = DefinitionDynamicConstraint(
                        exp_t='{0}_p[t] == {1}[t] * {2}_p[t]'.format(
                            self.rev_unit_downstream.production_unit.name,
                            up2down_eff,
                            self.rev_unit_upstream.consumption_unit.name),
                        t_range='for t in time.I', name='conversion_up2down',
                        parent=self)
                else:
                    raise ValueError('The up2down efficiency values should be '
                                     'lower than 1')
            else:
                raise IndexError('The length of the up2down efficiency '
                                 'vector should be of the same length as the '
                                 'TimeUnit of the studied period')

        elif isinstance(up2down_eff, dict):  # up2down_eff is a dict of
            # values
            if len(up2down_eff) == self.time.LEN:
                if all(e <= 1 for e in up2down_eff.values()):
                    self.conversion_up2down = DefinitionDynamicConstraint(
                        exp_t='{0}_p[t] == {1}[t] * {2}_p[t]'.format(
                            self.rev_unit_downstream.production_unit.name,
                            up2down_eff.values(),
                            self.rev_unit_upstream.consumption_unit.name),
                        t_range='for t in time.I', name='conversion_up2down',
                        parent=self)
                else:
                    raise ValueError('The up2down efficiency values should be '
                                     'lower than 1')
            else:
                raise IndexError('The length of the up2down efficiency '
                                 'dictionary should be of the same length as '
                                 'the TimeUnit of the studied period')
        else:
            raise TypeError(
                "The up2down efficiency should be a mean value or a vector "
                "(list or dict) on the whole time period !")

        # Managing the downstream to upstream conversion for various types
        # of down2up_eff:

        if isinstance(down2up_eff, (int, float)):  # down2up_eff is a mean
            # value
            if down2up_eff <= 1:
                self.conversion_down2up = DefinitionDynamicConstraint(
                    exp_t='{0}_p[t] == {1} * {2}_p[t]'.format(
                        self.rev_unit_upstream.production_unit.name,
                        down2up_eff,
                        self.rev_unit_downstream.consumption_unit.name),
                    t_range='for t in time.I', name='conversion_down2up',
                    parent=self)
            else:
                raise ValueError('The down2up_efficiency should be lower '
                                 'than 1')

        elif isinstance(down2up_eff, list):  # down2up_eff is a list of
            # values
            if len(down2up_eff) == self.time.LEN:  # it must have the right
                # size, i.e. the TimeUnit length.
                if all(e <= 1 for e in down2up_eff):
                    self.conversion_down2up = DefinitionDynamicConstraint(
                        exp_t='{0}_p[t] == {1}[t] * {2}_p[t]'.format(
                            self.rev_unit_upstream.production_unit.name,
                            down2up_eff,
                            self.rev_unit_downstream.consumption_unit.name),
                        t_range='for t in time.I', name='conversion_down2up',
                        parent=self)
                else:
                    raise ValueError('The down2up efficiency values should be '
                                     'lower than 1')
            else:
                raise IndexError('The length of the down2up efficiency '
                                 'vector should be of the same length as the '
                                 'TimeUnit of the studied period')

        elif isinstance(down2up_eff, dict):  # down2up_eff is a dict of
            # values
            if len(down2up_eff) == self.time.LEN:
                if all(e <= 1 for e in down2up_eff.values()):
                    self.conversion_down2up = DefinitionDynamicConstraint(
                        exp_t='{0}_p[t] == {1}[t] * {2}_p[t]'.format(
                            self.rev_unit_upstream.production_unit.name,
                            down2up_eff.values(),
                            self.rev_unit_downstream.consumption_unit.name),
                        t_range='for t in time.I', name='conversion_down2up',
                        parent=self)
                else:
                    raise ValueError('The down2up efficiency values should be '
                                     'lower than 1')
            else:
                raise IndexError('The length of the down2up efficiency '
                                 'dictionary should be of the same length as '
                                 'the TimeUnit of the studied period')
        else:
            raise TypeError(
                "The down2up efficiency should be a mean value or a vector "
                "(list or dict) on the whole time period !")


class HeatPump(ConversionUnit):
    """
    **Description**

        Simple Heat Pump with an electricity consumption, a thermal production
        and a thermal consumption. It has a theoretical coefficient of
        performance COP and inherits from ConversionUnit.

    **Attributes**

     * thermal_production_unit : thermal production unit (condenser)
     * elec_consumption_unit : electricity consumption unit (electrical
       input)
     * thermal_consumption_unit : heay consumption unit (evaporator)
     * COP : Quantity describing the coefficient of performance of the
       heat pump
     * conversion : Definition Dynamic Constraint linking the electrical
     input to the thermal output through the electrical to thermal ratio
     * power_flow : Definition Dynamic constraint linking the thermal output
     to the electrical and thermal inputs in relation to the losses.

    """

    def __init__(self, time, name, pmin_in_elec=1e-5, pmax_in_elec=1e+5,
                 p_in_elec=None, pmin_in_therm=1e-5, pmax_in_therm=1e+5,
                 p_in_therm=None, pmin_out_therm=1e-5, pmax_out_therm=1e+5,
                 p_out_therm=None, cop=3, losses=0):
        """
        :param time: TimeUnit describing the studied time period
        :param name: name of the heat pump
        :param pmin_in_elec:  minimal incoming electrical power
        :param pmax_in_elec: maximal incoming electrical power
        :param p_in_elec: power input for the electrical consumption unit
        :param pmin_in_therm: minimal incoming thermal power
        :param pmax_in_therm: maximal incoming thermal power
        :param p_in_therm: power input for the thermal consumption unit
        :param pmin_out_therm: minimal power output (thermal)
        :param pmax_out_therm: maximal power output (thermal)
        :param p_out_therm: power output (thermal)
        :param cop: Coefficient Of Performance of the Heat Pump (cop>1)
        :param losses: losses as a percentage of thermal energy produced (p_out)
        """

        if p_out_therm is None:
            self.thermal_production_unit = VariableProductionUnit(
                time, name + '_therm_prod', energy_type='Thermal',
                p_min=pmin_out_therm, p_max=pmax_out_therm)
        else:
            self.thermal_production_unit = FixedProductionUnit(
                time, name + '_therm_prod', energy_type='Thermal',
                p=p_out_therm)

        if p_in_therm is None:
            self.thermal_consumption_unit = VariableConsumptionUnit(
                time, name + '_therm_cons', energy_type='Thermal',
                p_min=pmin_in_therm, p_max=pmax_in_therm)
        else:
            self.thermal_consumption_unit = FixedConsumptionUnit(
                time, name + '_therm_cons', energy_type='Thermal',
                p=p_in_therm)

        if p_in_elec is None:
            self.elec_consumption_unit = VariableConsumptionUnit(
                time, name + '_elec_cons', p_min=pmin_in_elec,
                p_max=pmax_in_elec, energy_type='Electrical')
        else:
            self.elec_consumption_unit = FixedConsumptionUnit(
                time, name, p=p_in_elec, energy_type='Electrical')

        ConversionUnit.__init__(self, time, name,
                                prod_units=[self.thermal_production_unit],
                                cons_units=[self.thermal_consumption_unit,
                                            self.elec_consumption_unit])

        self.COP = Quantity(name='COP', value=cop, parent=self)

        if isinstance(self.COP.value, (int, float)):  # The cop has a single
            #  value
            if self.COP.value >= 1:  # The cop value should be greater than 1
                self.conversion = DefinitionDynamicConstraint(
                    exp_t='{0}_p[t] == {1} * {2}_p[t]'.format(
                        self.thermal_production_unit.name,
                        self.COP.value,
                        self.elec_consumption_unit.name),
                    t_range='for t in time.I', name='conversion', parent=self)

                self.power_flow = DefinitionDynamicConstraint(
                    exp_t='{0}_p[t]*(1+{1}) == {2}_p[t] + {3}_p[t]'
                        .format(self.thermal_production_unit.name, losses,
                                self.thermal_consumption_unit.name,
                                self.elec_consumption_unit.name),
                    t_range='for t in time.I',
                    name='power_flow', parent=self)
            else:
                raise ValueError("The COP value should be greater than 1")

        elif isinstance(self.COP.value, list):  # The cop has a list of values
            if len(self.COP.value) == self.time.LEN:
                if all(c >= 1 for c in self.COP.value):
                    self.conversion = DefinitionDynamicConstraint(
                        exp_t='{0}_p[t] == {1}[t] * {2}_p[t]'.format(
                            self.thermal_production_unit.name,
                            self.COP.value,
                            self.elec_consumption_unit.name),
                        t_range='for t in time.I', name='conversion',
                        parent=self)
                    self.power_flow = DefinitionDynamicConstraint(
                        exp_t='{0}_p[t]*(1+{1}) == {2}_p[t] + {3}_p[t]'
                            .format(self.thermal_production_unit.name, losses,
                                    self.thermal_consumption_unit.name,
                                    self.elec_consumption_unit.name),
                        t_range='for t in time.I',
                        name='power_flow', parent=self)
                else:
                    raise ValueError("The COP values should be greater than 1")
            else:
                raise IndexError("The COP should have the same length as the "
                                 "studied time period")

        elif isinstance(self.COP.value, dict):  # The cop has a dict
            # referencing its values.
            if len(self.COP.value) == self.time.LEN:
                if all(c >= 1 for c in self.COP.value.values()):
                    self.conversion = DefinitionDynamicConstraint(
                        exp_t='{0}_p[t] == {1}[t] * {2}_p[t]'.format(
                            self.thermal_production_unit.name,
                            self.COP.value,
                            self.elec_consumption_unit.name),
                        t_range='for t in time.I', name='conversion',
                        parent=self)
                    self.power_flow = DefinitionDynamicConstraint(
                        exp_t='{0}_p[t]*(1+{1}) == {2}_p[t] + {3}_p[t]'
                            .format(self.thermal_production_unit.name, losses,
                                    self.thermal_consumption_unit.name,
                                    self.elec_consumption_unit.name),
                        t_range='for t in time.I',
                        name='power_flow', parent=self)
                else:
                    raise ValueError("The COP values should be greater than 1")
            else:
                raise IndexError("The COP should have the same length as the "
                                 "studied time period")
        else:
            raise TypeError(
                "The assigned cop should be a mean value or a vector "
                "(dict or list) over the studied time period !")



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module define the main Core object on which the energy units and actors
will be based**

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from .elements import Quantity, Constraint, Objective, \
    TechnicalConstraint, TechnicalDynamicConstraint

__docformat__ = "restructuredtext en"


class OptObject:
    """
    **Description**

        OptObject class is used as an "abstract class", i.e. it defines some
        general attributes and methods but doesn't contain variable,
        constraint nor objective. In the OMEGAlpes package, all the
        subsystem models are represented by a unit. A model is then
        generated adding OptObject to it.
        Variable, objective and constraints declarations are usually done
        using the __init__ method of the OptObject class.

    **Attributes**

     - name
     - description
     - _quantities_list : list of the quanitities of the OptObject
       (active or not)
     - _constraints_list : list of the constraints of the OptObject(active or
       not)
     - _technical_constraints_list : list of the constraints of the OptObject
       (active or not)
     - _objectives_list : list of the objectives of the OptObject
       (active or not)


    **Methods**

     - __str__: defines the
     - __repr__: defines the unit with its name
     - get_constraints_list
     - get_constraints_name_list
     - get_external_constraints_list
     - get_external_constraints_name_list
     - get_objectives_list
     - get_objectives_name_list
     - get_quantities_list
     - get_quantities_name_list
     - deactivate_optobject_external_constraints


    .. note::
        The OptObject class shouldn't be instantiated in a python script,
        except if you want to create your own model from the beginning.
        In this case, one should consider creating a new class
        NewModel(OptObject).
    """

    def __init__(self, name='U0', description="Optimization object",
                 verbose=True):
        self.name = name
        self.description = description
        self.units_list = []
        self._quantities_list = []
        self._constraints_list = []  # all constraints are in this list
        self._technical_constraints_list = []  # only for technical constraints
        self._actor_constraints_list = []  # only for actor's constraints
        self._objectives_list = []

        if verbose:
            print(("Creating the {0}.".format(name)))

    def __str__(self):
        """"
        Add in the expression of the unit the variables, constraints and
        objectives

        :return: string
        """
        import numpy
        var = {}
        cst = {}
        cstr = {}
        obj = {}
        exp = '<OMEGALPES.general.units.OptObject: \nname: {0} \ndescription: {1}' \
              '\n'.format(self.name, self.description)
        for u_key in list(self.__dict__.keys()):
            key: (Quantity, Constraint, Objective) = getattr(self, u_key)
            if isinstance(key, Quantity):
                if isinstance(key.opt, bool):
                    if key.opt:
                        var[u_key] = key
                    else:
                        cst[u_key] = key
                elif isinstance(key.opt, dict):
                    if numpy.array(list(key.opt.values())).all():
                        var[u_key] = key
                    else:
                        cst[u_key] = key
            elif isinstance(key, Constraint):
                cstr[u_key] = key
            elif isinstance(key, Objective):
                obj[u_key] = key
        exp += '\nOptimization variables:\n'
        for u_key in list(var.keys()):
            exp += 'name: ' + getattr(self, u_key).name + '\n'
        exp += '\nConstants:\n'
        for u_key in list(cst.keys()):
            exp += 'name: ' + getattr(self, u_key).name + ',  value: ' + \
                   str(getattr(self, u_key).value) + '\n'
        exp += '\nConstraints:\n'
        for u_key in list(cstr.keys()):
            exp += '[' + str(getattr(self, u_key).active) + ']' + ' name: ' + \
                   getattr(self, u_key).name + ' exp: ' + \
                   str(getattr(self, u_key).exp) + '\n'
        exp += '\nObjective:\n'
        for u_key in list(obj.keys()):
            exp += '[' + str(getattr(self, u_key).active) + ']' + 'name: ' \
                   + getattr(self, u_key).name + ' exp: ' + \
                   str(getattr(self, u_key).exp) + '\n'
        return exp

    def __repr__(self):
        """
        Return the description of the unit considering the name

        :return: string
        """
        return "<OMEGALPES.general.optimisation.units.OptObject: name:\'{0}\'>" \
            .format(self.name)

    def get_constraints_list(self):
        """ Get the constraints associated with the unit
        as a dictionary shape ['constraint_name' , constraint]
        """
        cst_dict = {}
        for cst in self._constraints_list:
            cst_dict[cst.name] = cst
        return cst_dict

    def get_constraints_name_list(self):
        """ Get the names of the constraints associated with the unit """
        constraints_name_list = []
        for constraint in self._constraints_list:
            constraints_name_list.append(constraint.name)
        return constraints_name_list

    def get_technical_constraints_list(self):
        """ Get the technical constraints associated with the unit
        as a dictionary shape ['constraint_name' , constraint]
        """
        tech_cst_dict = {}
        for tech_cst in self._technical_constraints_list:
            tech_cst_dict[tech_cst.name] = tech_cst
        return tech_cst_dict

    def get_technical_constraints_name_list(self):
        """ Get the names of the external constraints associated with the
        unit """
        tech_csts_name_list = []
        for tech_cst in self._technical_constraints_list:
            tech_csts_name_list.append(tech_cst.name)
        return tech_csts_name_list

    def get_actor_constraints_list(self):
        """ Get the technical constraints associated with the unit
        as a dictionary shape ['constraint_name' , constraint]
        """
        actor_cst_dict = {}
        for actor_cst in self._actor_constraints_list:
            actor_cst_dict[actor_cst.name] = actor_cst
        return actor_cst_dict

    def get_actor_constraints_name_list(self):
        """ Get the names of the external constraints associated with the
        unit """
        actor_csts_name_list = []
        for actor_cst in self._actor_constraints_list:
            actor_csts_name_list.append(actor_cst.name)
        return actor_csts_name_list

    def get_objectives_list(self):
        """ Get objectives associated with the unit
        as a dictionary shape ['objective_name' , objective]
        """
        obj_dict = {}
        for obj in self._objectives_list:
            obj_dict[obj.name] = obj
        return obj_dict

    def get_objectives_name_list(self):
        """ Get the names of the objectives associated with the unit """
        objectives_name_list = []
        for objective in self._objectives_list:
            objectives_name_list.append(objective.name)
        return objectives_name_list

    def get_quantities_list(self):
        """ Get the quantities associated with the unit
        as a dictionary shape ['quantity_name' , quantity]
        """
        quantity_dict = {}
        for quantity in self._quantities_list:
            quantity_dict[quantity.name] = quantity
        return quantity_dict

    def get_quantities_name_list(self):
        """ Get the names of the quantities associated with the unit """
        quantities_name_list = []
        for quantity in self._quantities_list:
            quantities_name_list.append(quantity.name)
        return quantities_name_list

    def deactivate_optobject_external_constraints(self, ext_cst_name_list=None):
        """
        Enable to remove an external constraint linked with an OptObject

        :param ext_cst_name_list: list of external constraint that would be
            removed
        """
        if not ext_cst_name_list:
            cst_dict = self.get_constraints_list()
            for key in cst_dict:
                if isinstance(cst_dict[key],
                              TechnicalConstraint or
                              TechnicalDynamicConstraint):
                    cst_dict[key].deactivate_constraint()
                else:
                    Warning('Only external constraints can be removed')
        else:
            cst_dict = self.get_constraints_list()
            for cst_name in ext_cst_name_list:
                for key in cst_dict:
                    if cst_name == key:
                        if isinstance(cst_dict[key],
                                      TechnicalConstraint or TechnicalDynamicConstraint):
                            cst_dict[key].deactivate_constraint()
                        else:
                            Warning('Only external constraints can be removed')
                    else:
                        pass

    def _deactivate_optobject_constraints(self, cst_name_list=None):
        """
        /!\ Should only be used for debug
        Enable to remove all kind of constraints (internal and external) linked
        with an OptObject.
        The constraint can be selected if cst_name_list is filled
        Warning, up and low limits are not constraints

        :param cst_name_list: list of external constraint that would be
            removed
        """
        if not cst_name_list:
            cst_dict = self.get_constraints_list()
            for key in cst_dict:
                cst_dict[key]._deactivate_constraint()

        else:
            cst_dict = self.get_constraints_list()
            for cst_name in cst_name_list:
                for key in cst_dict:
                    if cst_name == key:
                        cst_dict[key]._deactivate_constraint()
                    else:
                        pass
